import astropy.units as u
import matplotlib.pyplot as plt
import mpl_animators
import numpy as np
import pytest
from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs import WCS

from ndcube.ndcube import NDCube
from ndcube.tests.helpers import figure_test
from ndcube.visualization import PlotterDescriptor


@figure_test
def test_plot_1D_cube(ndcube_1d_l):
    fig = plt.figure()
    ax = ndcube_1d_l.plot()
    assert isinstance(ax, WCSAxes)
    return fig


@figure_test
@pytest.mark.parametrize(("ndcube_4d", "cslice", "kwargs"),
                         (
                             ("ln_lt_l_t", np.s_[0, 0, 0, :], {}),
                             ("ln_lt_l_t", np.s_[0, 0, :, 0], {}),
                             ("ln_lt_l_t", np.s_[0, :, 0, 0], {}),
                             ("ln_lt_l_t", np.s_[:, 0, 0, 0], {}),

                             ("uncertainty", np.s_[0, 0, 0, :], {}),
                             ("unit_uncertainty", np.s_[0, 0, 0, :], {'data_unit': u.mJ}),

                             ("mask", np.s_[0, 0, 0, :], {'marker': 'o'}),),
                         indirect=["ndcube_4d"])
def test_plot_1D_cube_from_slice(ndcube_4d, cslice, kwargs):
    # TODO: The output for the spatial plots is inconsistent between the lat
    # slice and the lon slice.
    fig = plt.figure()

    sub = ndcube_4d[cslice]
    ax = sub.plot(**kwargs)
    assert isinstance(ax, WCSAxes)

    return fig


@figure_test
def test_plot_2D_cube(ndcube_2d_ln_lt):
    fig = plt.figure()
    ax = ndcube_2d_ln_lt.plot()
    assert isinstance(ax, WCSAxes)
    return fig


@figure_test
def test_plot_2D_cube_colorbar(ndcube_2d_ln_lt):
    fig = plt.figure()
    ax = ndcube_2d_ln_lt.plot()
    assert isinstance(ax, WCSAxes)
    plt.colorbar()
    return fig


@figure_test
def test_plot_2D_cube_custom_axis(ndcube_2d_ln_lt):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection=ndcube_2d_ln_lt.wcs)
    ndcube_2d_ln_lt.plot(axes=ax)
    return fig


@figure_test
def test_plot_2D_cube_custom_axis_plot_axes(ndcube_2d_ln_lt):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection=ndcube_2d_ln_lt.wcs)
    ndcube_2d_ln_lt.plot(axes=ax, plot_axes=('x', 'y'))
    return fig


@figure_test
@pytest.mark.parametrize(("ndcube_4d", "cslice", "kwargs"),
                         (
                             ("ln_lt_l_t", np.s_[0, 0, :, :], {}),
                             ("ln_lt_l_t", np.s_[0, :, :, 0], {}),
                             ("ln_lt_l_t", np.s_[:, :, 0, 0], {}),
                             ("unit_uncertainty", np.s_[0, 0, :, :], {'data_unit': u.mJ}),
                             ("mask", np.s_[0, :, 0, :], {}),),
                         indirect=["ndcube_4d"])
def test_plot_2D_cube_from_slice(ndcube_4d, cslice, kwargs):
    fig = plt.figure()

    sub = ndcube_4d[cslice]
    ax = sub.plot(**kwargs)
    assert isinstance(ax, WCSAxes)

    return fig


@figure_test
def test_animate_2D_cube(ndcube_2d_ln_lt):
    cube = ndcube_2d_ln_lt
    ax = cube.plot(plot_axes=[None, 'x'])
    assert isinstance(ax, mpl_animators.ArrayAnimatorWCS)

    return ax.fig


@figure_test
@pytest.mark.parametrize(("ndcube_4d", "cslice", "kwargs"),
                         (
                             ("ln_lt_l_t", np.s_[:, :, 0, :], {}),
                             ("ln_lt_l_t", np.s_[:, :, 0, :], {'plot_axes': [..., 'x']}),
                             ("ln_lt_l_t", None, {}),
                             ("ln_lt_l_t", None, {"plot_axes": [0, 0, 'x', 'y'], "axes_units": [None, None, u.pm, None]}),
                             ("ln_lt_l_t", None, {"plot_axes": [0, 'x', 0, 'y']}),
                             ("ln_lt_l_t", np.s_[0, :, :, :], {}),
                             ("ln_lt_l_t", np.s_[:, :, :, :], {}),
                             ("unit_uncertainty", np.s_[0, :, :, :], {'data_unit': u.mJ}),
                             ("mask", np.s_[:, :, :, :], {}),),
                         indirect=["ndcube_4d"])
def test_animate_cube_from_slice(ndcube_4d, cslice, kwargs):
    if cslice:
        sub = ndcube_4d[cslice]
    else:
        sub = ndcube_4d
    ax = sub.plot(**kwargs)
    assert isinstance(ax, mpl_animators.ArrayAnimatorWCS)

    return ax.fig


@pytest.mark.parametrize(("ndcube_4d", "cslice"),
                         [("ln_lt_l_t", np.s_[:, :, 0, 0])], indirect=["ndcube_4d"])
def test_mpl_axes(ndcube_4d, cslice):
    ndcube_2d = ndcube_4d[cslice]
    ax = plt.subplot(projection=ndcube_2d)
    assert isinstance(ax, WCSAxes)
    plt.close()


def test_plotter_is_None(ndcube_1d_l):
    class NewCube(NDCube):
        plotter = PlotterDescriptor(default_type=None)

    cube = NewCube(np.zeros((1, 1)), wcs=WCS(naxis=2))
    assert cube.plotter is None

    with pytest.raises(NotImplementedError, match="no default plotting functionality is available"):
        cube.plot()

    # You can't (and shouldn't) set the plotter to None unless it's done at
    # descriptor init time:
    with pytest.raises(TypeError):
        ndcube_1d_l.plotter = None



import abc
from typing import Any, Tuple, Union, Iterable
from numbers import Integral
from functools import reduce, partial

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.wcsapi import BaseHighLevelWCS
from astropy.wcs.wcsapi.wrappers.sliced_wcs import SlicedLowLevelWCS, sanitize_slices

from ndcube.utils.wcs import convert_between_array_and_pixel_axes
from ndcube.wcs.wrappers import CompoundLowLevelWCS

from .table_coord import (BaseTableCoordinate, MultipleTableCoordinate, QuantityTableCoordinate,
                          SkyCoordTableCoordinate, TimeTableCoordinate)

__all__ = ['ExtraCoords']


class ExtraCoordsABC(abc.ABC):
    """
    A representation of additional world coordinates associated with pixel axes.

    ExtraCoords can be initialised by either specifying a
    `~astropy.wcs.wcsapi.LowLevelWCS` object and a ``mapping``, or it can be
    built up by specifying one or more lookup tables.

    Parameters
    ----------
    wcs
        The WCS specifying the extra coordinates.
    mapping
       The mapping between the array dimensions and pixel dimensions in the
       extra coords object. This is an iterable of ``(array_dimension, pixel_dimension)`` pairs
       of length equal to the number of pixel dimensions in the extra coords.

    """
    @abc.abstractmethod
    def add(self,
            name: Union[str, Iterable[str]],
            array_dimension: Union[int, Iterable[int]],
            lookup_table: Any,
            physical_types: Union[str, Iterable[str]] = None,
            **kwargs):
        """
        Add a coordinate to this ``ExtraCoords`` based on a lookup table.

        Parameters
        ----------
        name : `str` or sequence of `str`
            The name(s) for these world coordinate(s).
        array_dimension : `int` or `tuple` of `int`
            The pixel dimension(s), in the array, to which this lookup table corresponds.
        lookup_table : `object` or sequence of `object`
            The lookup table. A `BaseTableCoordinate <.table_coord>` subclass or anything
            that can instantiate one, i.e. currently a `~astropy.time.Time`,
            `~astropy.coordinates.SkyCoord`, or a (sequence of) `~astropy.units.Quantity`.
        physical_types: `str` or iterable of `str`, optional
            Descriptor(s) of the `physical type <../data_classes.html#dimensions-and-physical-types>`_
            associated with each axis; length must match the number of dimensions in
            ``lookup_table``.
        """

    @abc.abstractmethod
    def keys(self) -> Iterable[str]:
        """
        The world axis names for all the coordinates in the extra coords.
        """

    @property
    @abc.abstractmethod
    def mapping(self) -> Iterable[Tuple[int, int]]:
        """
        The mapping between the array dimensions and pixel dimensions.

        This is an iterable of ``(array_dimension, pixel_dimension)`` pairs
        of length equal to the number of pixel dimensions in the extra coords.
        """

    @property
    @abc.abstractmethod
    def wcs(self) -> BaseHighLevelWCS:
        """
        A WCS object representing the world coordinates described by this ``ExtraCoords``.

        .. note::
            This WCS object does not map to the pixel dimensions of the array
            associated with the `.NDCube` object. It has the number of
            pixel dimensions equal to the number of inputs to the transforms to
            get the world coordinates (normally equal to the number of world
            coordinates). Therefore using this WCS directly might lead to some
            confusing results.

        """

    @abc.abstractmethod
    def __getitem__(self, item: Union[str, int, slice, Iterable[Union[str, int, slice]]]) -> "ExtraCoordsABC":
        """
        ExtraCoords can be sliced with either a string, or a numpy like slice.

        When sliced with a string it should return a new ExtraCoords object
        with only those coordinates with the given names. When sliced with a
        numpy array like slice it should return a new ExtraCoords with the
        slice applied. Supporting step is not required and "fancy indexing" is
        not supported.
        """


class ExtraCoords(ExtraCoordsABC):
    """
    A representation of additional world coordinates associated with pixel axes.

    ExtraCoords can be initialised by either specifying a
    `~astropy.wcs.wcsapi.LowLevelWCS` object and a ``mapping``, or it can be
    built up by specifying one or more lookup tables.

    Parameters
    ----------
    wcs
        The WCS specifying the extra coordinates.
    mapping
       The mapping between the array dimensions and pixel dimensions in the
       extra coords object. This is an iterable of ``(array_dimension, pixel_dimension)`` pairs
       of length equal to the number of pixel dimensions in the extra coords.

    """
    def __init__(self, ndcube=None):
        super().__init__()

        # Setup private attributes
        self._wcs = None
        self._mapping = None

        # Lookup tables is a list of (pixel_dim, LookupTableCoord) to allow for
        # one pixel dimension having more than one lookup coord.
        self._lookup_tables = list()
        self._dropped_tables = list()

        # We need a reference to the parent NDCube
        self._ndcube = ndcube

    @classmethod
    def from_lookup_tables(cls, names, pixel_dimensions, lookup_tables, physical_types=None):
        """
        Construct a new ExtraCoords instance from lookup tables.

        This is a convience wrapper around `.add` which does not
        expose all the options available in that method.

        Parameters
        ----------
        names : `tuple` of `str`
            The names of the world coordinates.
        pixel_dimensions : `tuple` of `int`
            The pixel dimensions (in the array) to which the ``lookup_tables``
            apply. Must be the same length as ``lookup_tables``.
        lookup_tables : iterable of `object`
            The lookup tables which specify the world coordinates for the ``pixel_dimensions``.
            Must be `BaseTableCoordinate <.table_coord>` subclass instances or objects from
            which to instantiate them (see `.ExtraCoords.add`).
        physical_types: sequence of `str` or of sequences of `str`, optional
            Descriptors of the `physical types <../data_classes.html#dimensions-and-physical-types>`_
            associated with each axis in the tables. Must be the same length as ``lookup_tables``;
            and length of each element must match the number of dimensions in corresponding
            ``lookup_tables[i]``.

        Returns
        -------
        `ndcube.extra_coords.ExtraCoords`

        """
        if len(pixel_dimensions) != len(lookup_tables):
            raise ValueError(
                "The length of pixel_dimensions and lookup_tables must match."
            )

        if physical_types is None:
            physical_types = len(lookup_tables) * [physical_types]
        elif len(physical_types) != len(lookup_tables):
            raise ValueError("The number of physical types and lookup_tables must match.")

        extra_coords = cls()

        for name, pixel_dim, lookup_table, physical_type in zip(names, pixel_dimensions,
                                                                lookup_tables, physical_types):
            extra_coords.add(name, pixel_dim, lookup_table, physical_types=physical_type)

        return extra_coords

    def add(self, name, array_dimension, lookup_table, physical_types=None, **kwargs):
        # docstring in ABC

        if self._wcs is not None:
            raise ValueError(
                "Can not add a lookup_table to an ExtraCoords which was instantiated with a WCS object."
            )

        kwargs['names'] = [name] if not isinstance(name, (list, tuple)) else name

        if isinstance(lookup_table, BaseTableCoordinate):
            coord = lookup_table
        elif isinstance(lookup_table, Time):
            coord = TimeTableCoordinate(lookup_table, physical_types=physical_types, **kwargs)
        elif isinstance(lookup_table, SkyCoord):
            coord = SkyCoordTableCoordinate(lookup_table, physical_types=physical_types, **kwargs)
        elif isinstance(lookup_table, (list, tuple)):
            coord = QuantityTableCoordinate(*lookup_table, physical_types=physical_types, **kwargs)
        elif isinstance(lookup_table, u.Quantity):
            coord = QuantityTableCoordinate(lookup_table, physical_types=physical_types, **kwargs)
        else:
            raise TypeError(f"The input type {type(lookup_table)} isn't supported")

        self._lookup_tables.append((array_dimension, coord))

        # Sort the LUTs so that the mapping and the wcs are ordered in pixel dim order
        self._lookup_tables = list(sorted(self._lookup_tables,
                                          key=lambda x: x[0] if isinstance(x[0], int) else x[0][0]))

    @property
    def _name_lut_map(self):
        """
        Map of world names to the corresponding `.LookupTableCoord`
        """
        return {lut[1].wcs.world_axis_names: lut for lut in self._lookup_tables}

    def keys(self):
        # docstring in ABC
        if not self.wcs:
            return tuple()

        return tuple(self.wcs.world_axis_names) if self.wcs.world_axis_names else None

    @property
    def mapping(self):
        # docstring in ABC
        if self._mapping:
            return self._mapping

        # If mapping is not set but lookup_tables is empty then the extra
        # coords is empty, so there is no mapping.
        if not self._lookup_tables:
            return tuple()

        # The mapping is from the array index (position in the list) to the
        # pixel dimensions (numbers in the list)
        lts = [list([lt[0]] if isinstance(lt[0], Integral) else lt[0]) for lt in self._lookup_tables]
        converter = partial(convert_between_array_and_pixel_axes, naxes=len(self._ndcube.dimensions))
        pixel_indicies = [list(converter(np.array(ids))) for ids in lts]
        return tuple(reduce(list.__add__, pixel_indicies))

    @mapping.setter
    def mapping(self, mapping):
        if self._mapping is not None:
            raise AttributeError("Can't set mapping if a mapping has already been specified.")

        if self._lookup_tables:
            raise AttributeError(
                "Can't set mapping manually when ExtraCoords is built from lookup tables."
            )

        if self._wcs is not None:
            if not max(mapping) <= self._wcs.pixel_n_dim - 1:
                raise ValueError(
                    "Values in the mapping can not be larger than the number of pixel dimensions in the WCS."
                )

        self._mapping = mapping

    @property
    def wcs(self):
        # docstring in ABC
        if self._wcs is not None:
            return self._wcs

        if not self._lookup_tables:
            return None

        tcoords = set(lt[1] for lt in self._lookup_tables)
        # created a sorted list of unique items
        _tmp = set()  # a temporary set
        tcoords = [x[1] for x in self._lookup_tables if x[1] not in _tmp and _tmp.add(x[1]) is None]
        return MultipleTableCoordinate(*tcoords).wcs

    @wcs.setter
    def wcs(self, wcs):
        if self._wcs is not None:
            raise AttributeError(
                "Can't set wcs if a WCS has already been specified."
            )

        if self._lookup_tables:
            raise AttributeError(
                "Can't set wcs manually when ExtraCoords is built from lookup tables."
            )

        if self._mapping is not None:
            if not max(self._mapping) <= wcs.pixel_n_dim - 1:
                raise ValueError(
                    "Values in the mapping can not be larger than the number of pixel dimensions in the WCS."
                )

        self._wcs = wcs

    def _getitem_string(self, item):
        """
        Slice the Extracoords based on axis names.
        """

        for names, lut in self._name_lut_map.items():
            if item in names:
                new_ec = ExtraCoords(ndcube=self._ndcube)
                new_ec._lookup_tables = [lut]
                return new_ec

        raise KeyError(f"Can't find the world axis named {item} in this ExtraCoords object.")

    def _getitem_lookup_tables(self, item):
        """
        Apply an array slice to the lookup tables.

        Returns a new ExtraCoords object with modified lookup tables.
        """
        dropped_tables = set()
        new_lookup_tables = set()
        ndims = max([lut[0] if isinstance(lut[0], Integral) else max(lut[0])
                     for lut in self._lookup_tables]) + 1
        # Determine how many dimensions will be dropped by slicing below each dimension.
        if isinstance(item, Integral):
            n_dropped_dims = np.ones(ndims, dtype=int)
            item = tuple([item] + [slice(None)] * (ndims - 1))
        elif isinstance(item, slice):
            n_dropped_dims = np.zeros(ndims, dtype=int)
            item = tuple([item] + [slice(None)] * (ndims - 1))
        else:
            item = list(item) + [slice(None)] * (ndims - len(item))
            n_dropped_dims = np.cumsum([isinstance(i, Integral) for i in item])
        for lut_axis, lut in self._lookup_tables:
            lut_axes = (lut_axis,) if not isinstance(lut_axis, tuple) else lut_axis
            new_lut_axes = tuple(ax - n_dropped_dims[ax] for ax in lut_axes)
            lut_slice = tuple(item[i] for i in lut_axes)
            if isinstance(lut_slice, tuple) and len(lut_slice) == 1:
                lut_slice = lut_slice[0]

            sliced_lut = lut[lut_slice]

            if sliced_lut.is_scalar():
                dropped_tables.add(sliced_lut)
            else:
                new_lookup_tables.add((new_lut_axes, sliced_lut))
        new_extra_coords = type(self)()
        new_extra_coords._lookup_tables = list(new_lookup_tables)
        new_extra_coords._dropped_tables = list(dropped_tables)
        return new_extra_coords

    def _getitem_wcs(self, item):
        item = sanitize_slices(item, self.wcs.pixel_n_dim)

        # It's valid to slice down the EC such that there is nothing left,
        # which is not a valid way to slice the WCS
        if len(item) == self.wcs.pixel_n_dim and all(isinstance(i, Integral) for i in item):
            return type(self)()

        subwcs = self.wcs[item]

        new_mapping = [self.mapping[i] for i, subitem in enumerate(item) if not isinstance(subitem, Integral)]

        new_ec = type(self)()
        new_ec.wcs = subwcs
        new_ec.mapping = new_mapping
        return new_ec

    def __getitem__(self, item):
        # docstring in ABC
        if isinstance(item, str):
            return self._getitem_string(item)

        if self._wcs:
            return self._getitem_wcs(item)

        elif self._lookup_tables:
            return self._getitem_lookup_tables(item)

        # If we get here this object is empty, so just return an empty extra coords
        # This is done to simplify the slicing in NDCube
        return self

    @property
    def dropped_world_dimensions(self):
        """
        Return an APE-14 like representation of any sliced out world dimensions.
        """

        if self._wcs:
            if isinstance(self._wcs, SlicedLowLevelWCS):
                return self._wcs.dropped_world_dimensions

        if self._lookup_tables or self._dropped_tables:
            mtc = MultipleTableCoordinate(*[lt[1] for lt in self._lookup_tables])
            mtc._dropped_coords = self._dropped_tables

            return mtc.dropped_world_dimensions

        return dict()

    @property
    def cube_wcs(self):
        """Produce a WCS that describes the associated NDCube with just the extra coords.

        For NDCube pixel axes without any extra coord, dummy axes are inserted.
        """
        wcses = [self.wcs]
        mapping = list(self.mapping)
        dummy_axes = self._cube_array_axes_without_extra_coords
        n_dummy_axes = len(dummy_axes)
        if n_dummy_axes > 0:
            dummy_wcs = WCS(naxis=n_dummy_axes)
            dummy_wcs.wcs.crpix = [1] * n_dummy_axes
            dummy_wcs.wcs.cdelt = [1] * n_dummy_axes
            dummy_wcs.wcs.crval = [0] * n_dummy_axes
            dummy_wcs.wcs.ctype = ["PIXEL"] * n_dummy_axes
            dummy_wcs.wcs.cunit = ["pix"] * n_dummy_axes
            wcses.append(dummy_wcs)
            mapping += list(dummy_axes)
        return CompoundLowLevelWCS(*wcses, mapping=mapping)

    @property
    def _cube_array_axes_without_extra_coords(self):
        """Return the array axes not associated with any extra coord."""
        return set(range(len(self._ndcube.dimensions))) - set(self.mapping)

    def __str__(self):
        classname = self.__class__.__name__
        elements = [f"{', '.join(table.names)} ({axes}) {table.physical_types}: {table}"
                    for axes, table in self._lookup_tables]
        length = len(classname) + 2 * len(elements) + sum(len(e) for e in elements)
        if length > np.get_printoptions()['linewidth']:
            joiner = ',\n ' + len(classname) * ' '
        else:
            joiner = ', '

        return f"{classname}({joiner.join(elements)})"

    def __repr__(self):
        return f"{object.__repr__(self)}\n{self}"



#
# Configuration file for the Sphinx documentation builder.

# -- Project information -----------------------------------------------------

project = 'ndcube'
copyright = '2021, The SunPy Community'
author = 'The SunPy Community'

# The full version, including alpha/beta/rc tags
from ndcube import __version__

release = __version__
is_development = '.dev' in __version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'sphinx.ext.mathjax',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.smart_resolver',
    'ndcube.utils.sphinx.code_context',
    'sphinx_changelog',
    'pytest_doctestplus.sphinx.doctestplus',
]

# Define code to be run by doctest in front of every code block.

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The reST default role (used for this markup: `text`) to use for all
# documents. Set to the "smart" one.
default_role = 'obj'

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/',
               (None, 'http://data.astropy.org/intersphinx/python3.inv')),
    'numpy': ('https://docs.scipy.org/doc/numpy/',
              (None, 'http://data.astropy.org/intersphinx/numpy.inv')),
    'matplotlib': ('https://matplotlib.org/',
                   (None, 'http://data.astropy.org/intersphinx/matplotlib.inv')),
    'astropy': ('http://docs.astropy.org/en/stable/', None),
    'sunpy': ('https://docs.sunpy.org/en/stable/', None),
    'mpl_animators': ('https://docs.sunpy.org/projects/mpl-animators/en/stable/', None),
    'gwcs': ('https://gwcs.readthedocs.io/en/stable/', None)}

# -- Options for HTML output -------------------------------------------------
# Load the theme config from the theme package
from sunpy_sphinx_theme.conf import *

# Use the ndcube logo rather than the sunpy one
html_logo = png_icon = 'logo/ndcube.png'
html_favicon = 'logo/favicon.png'

# Render inheritance diagrams in SVG
graphviz_output_format = 'svg'

graphviz_dot_args = [
    '-Nfontsize=10',
    '-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Efontsize=10',
    '-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Gfontsize=10',
    '-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif'
]



from unittest.mock import MagicMock

import astropy.units as u
import gwcs
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.time import Time, TimeDelta
from astropy.wcs import WCS

from ndcube import NDCube
from ndcube.extra_coords import ExtraCoords

# Fixtures


@pytest.fixture
def time_lut():
    return Time(["2011-01-01T00:00:00",
                 "2011-01-01T00:00:10",
                 "2011-01-01T00:00:20",
                 "2011-01-01T00:00:30"], format="isot")


@pytest.fixture
def wave_lut():
    return range(10, 20) * u.nm


@pytest.fixture
def skycoord_1d_lut():
    return SkyCoord(range(10), range(10), unit=u.deg)


@pytest.fixture
def skycoord_2d_lut():
    data = np.arange(9).reshape(3, 3), np.arange(9, 18).reshape(3, 3)
    return SkyCoord(*data, unit=u.deg)


@pytest.fixture
def quantity_2d_lut():
    ec_shape = (3, 3)
    return np.arange(np.product(ec_shape)).reshape(ec_shape) * u.m / u.s


# ExtraCoords from WCS


def test_empty_ec(wcs_1d_l):
    ec = ExtraCoords()
    # Test slice of an empty EC
    assert ec[0].wcs is None

    assert ec.mapping == tuple()
    assert ec.wcs is None
    assert ec.keys() == tuple()

    ec.wcs = wcs_1d_l
    assert ec.wcs is wcs_1d_l
    ec.mapping = (0,)
    assert ec.mapping == (0,)


def test_exceptions(wcs_1d_l):
    # Test unable to specify inconsistent dimensions and tables
    with pytest.raises(ValueError):
        ExtraCoords.from_lookup_tables(None, (0,), (0, 0))

    # Test unable to add to WCS EC
    ec = ExtraCoords()
    ec.wcs = wcs_1d_l
    ec.mapping = (0,)

    with pytest.raises(ValueError):
        ec.add(None, 0, None)

    with pytest.raises(KeyError):
        ExtraCoords()['empty']


def test_mapping_setter(wcs_1d_l, wave_lut):
    ec = ExtraCoords()
    ec.wcs = wcs_1d_l
    ec.mapping = (0,)

    with pytest.raises(AttributeError):
        ec.mapping = None

    ec = ExtraCoords()
    ec.wcs = wcs_1d_l
    with pytest.raises(ValueError):
        ec.mapping = (1,)

    ec = ExtraCoords()
    ec.add("wave", (0,), wave_lut)
    with pytest.raises(AttributeError):
        ec.mapping = None


def test_wcs_setter(wcs_1d_l, wave_lut):
    ec = ExtraCoords()
    ec.wcs = wcs_1d_l
    ec.mapping = (0,)

    with pytest.raises(AttributeError):
        ec.wcs = None

    ec = ExtraCoords()
    ec.mapping = (1,)
    with pytest.raises(ValueError):
        ec.wcs = wcs_1d_l

    ec = ExtraCoords()
    ec.add("wave", (0,), wave_lut)
    with pytest.raises(AttributeError):
        ec.wcs = None


def test_wcs_1d(wcs_1d_l):
    ec = ExtraCoords()
    ec.wcs = wcs_1d_l
    ec.mapping = (0,)

    assert ec.keys() == ('spectral',)
    assert ec.mapping == (0,)
    assert ec.wcs is wcs_1d_l

    subec = ec[1:]
    assert ec.keys() == ('spectral',)
    assert ec.mapping == (0,)
    assert np.allclose(ec.wcs.pixel_to_world_values(1), subec.wcs.pixel_to_world_values(1))

    subec = ec[0]
    assert subec.wcs is None


@pytest.fixture
def extra_coords_wave(wave_lut):
    cube = MagicMock()
    cube.dimensions = [10] * u.pix
    ec = ExtraCoords(cube)
    ec.add("wave", 0, wave_lut)

    return ec


# Extra Coords from lookup tables

# A single lookup along a dimension, i.e. Time along the second dim.
def test_single_from_lut(extra_coords_wave):
    ec = extra_coords_wave
    assert len(ec._lookup_tables) == 1
    assert ec.mapping == (0,)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert ec.wcs.pixel_n_dim == 1
    assert ec.wcs.world_n_dim == 1
    assert ec.wcs.world_axis_names == ("wave",)


def test_two_1d_from_lut(time_lut):
    cube = MagicMock()
    cube.dimensions = [10] * u.pix
    ec = ExtraCoords(cube)

    exposure_lut = range(10) * u.s
    ec.add("time", 0, time_lut)
    ec.add("exposure_time", 0, exposure_lut)
    assert len(ec._lookup_tables) == 2
    assert ec.mapping == (0, 0)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert ec.wcs.pixel_n_dim == 2
    assert ec.wcs.world_n_dim == 2
    assert ec.wcs.world_axis_names == ("time", "exposure_time")


def test_two_1d_from_lookup_tables(time_lut):
    """
    Create ExtraCoords from both tables at once using `from_lookup_tables` with `physical_types`.
    """

    exposure_lut = range(10) * u.s

    pt = ["custom:time:creation"]
    with pytest.raises(ValueError, match=r"The number of physical types and lookup_tables"):
        ec = ExtraCoords.from_lookup_tables(["time", "exposure_time"], (0, 0),
                                            [time_lut, exposure_lut], pt)

    pt.append("custom:time:duration")
    ec = ExtraCoords.from_lookup_tables(["time", "exposure_time"], (0, 0),
                                        [time_lut, exposure_lut], pt)

    # This has created an "orphan" extra_coords with no NDCube connected.
    with pytest.raises(AttributeError, match=r"'NoneType' object has no attribute 'dimensions'"):
        assert ec.mapping == (0, 0)

    assert len(ec._lookup_tables) == 2
    assert isinstance(ec.wcs, gwcs.WCS)
    assert ec.wcs.pixel_n_dim == 2
    assert ec.wcs.world_n_dim == 2
    assert ec.wcs.world_axis_names == ("time", "exposure_time")
    for i, physical_types in enumerate(pt):
        assert ec._lookup_tables[i][1].physical_types == [physical_types]


def test_skycoord(skycoord_1d_lut):
    cube = MagicMock()
    cube.dimensions = [10, 10] * u.pix

    ec = ExtraCoords(cube)
    ec.add(("lat", "lon"), (0, 1), skycoord_1d_lut, mesh=True)
    assert len(ec._lookup_tables) == 1
    assert ec.mapping == (1, 0)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert ec.wcs.pixel_n_dim == 2
    assert ec.wcs.world_n_dim == 2
    assert ec.wcs.world_axis_names == ("lat", "lon")


def test_skycoord_1_pixel(skycoord_1d_lut):
    cube = MagicMock()
    cube.dimensions = [10] * u.pix

    ec = ExtraCoords(cube)
    ec.add(("lon", "lat"), 0, skycoord_1d_lut, mesh=False)
    assert len(ec._lookup_tables) == 1
    assert ec.mapping == (0,)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert ec.wcs.pixel_n_dim == 1
    assert ec.wcs.world_n_dim == 2
    assert ec.wcs.world_axis_names == ("lon", "lat")

    sec = ec[1:4]
    assert sec.wcs.pixel_n_dim == 1
    assert sec.wcs.world_n_dim == 2
    assert sec.wcs.world_axis_names == ("lon", "lat")

    assert isinstance(sec.wcs.pixel_to_world(0), SkyCoord)


def test_skycoord_mesh_false(skycoord_2d_lut):
    cube = MagicMock()
    cube.dimensions = [10, 10] * u.pix

    ec = ExtraCoords(cube)
    ec.add(("lat", "lon"), (0, 1), skycoord_2d_lut, mesh=False)
    assert len(ec._lookup_tables) == 1
    assert ec.mapping == (1, 0)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert ec.wcs.pixel_n_dim == 2
    assert ec.wcs.world_n_dim == 2
    assert ec.wcs.world_axis_names == ("lat", "lon")


def test_extra_coords_index(skycoord_2d_lut, time_lut):
    cube = MagicMock()
    cube.dimensions = [10, 10] * u.pix

    ec = ExtraCoords(cube)
    ec.add(("lat", "lon"), (0, 1), skycoord_2d_lut, mesh=False)
    ec.add("exposure_time", (0,), time_lut)
    assert len(ec._lookup_tables) == 2
    assert ec.mapping == (1, 0, 1)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert ec.wcs.pixel_n_dim == 3
    assert ec.wcs.world_n_dim == 3
    assert ec.wcs.world_axis_names == ("lat", "lon", "exposure_time")

    sub_ec = ec["lon"]
    sub_ec._ndcube = cube
    assert len(sub_ec._lookup_tables) == 1
    assert sub_ec.mapping == (1, 0)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert sub_ec.wcs.pixel_n_dim == 2
    assert sub_ec.wcs.world_n_dim == 2
    assert sub_ec.wcs.world_axis_names == ("lat", "lon")

    sub_ec = ec["exposure_time"]
    sub_ec._ndcube = cube
    assert len(sub_ec._lookup_tables) == 1
    assert sub_ec.mapping == (1,)
    assert isinstance(ec.wcs, gwcs.WCS)
    assert sub_ec.wcs.pixel_n_dim == 1
    assert sub_ec.wcs.world_n_dim == 1
    assert sub_ec.wcs.world_axis_names == ("exposure_time",)


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_extra_coords_2d_quantity(quantity_2d_lut):
    ec = ExtraCoords()
    ec.add("velocity", (0, 1), quantity_2d_lut)

    ec.wcs.pixel_to_world(0, 0)

# Inspecting an extra coords
# Should be able to see what tables exists, what axes they account to, and what
# axes have missing dimensions.

# An additional spatial set (i.e. ICRS on top of HPC)


# Extra Coords with NDCube
def test_add_coord_after_create(time_lut):
    ndc = NDCube(np.random.random((10, 10)), wcs=WCS(naxis=2))
    assert isinstance(ndc.extra_coords, ExtraCoords)
    ndc.extra_coords.add("time", 0, time_lut)

    assert len(ndc.extra_coords._lookup_tables) == 1

    assert ndc.extra_coords["time"]._lookup_tables == ndc.extra_coords._lookup_tables


def test_combined_wcs(time_lut):
    ndc = NDCube(np.random.random((10, 10)), wcs=WCS(naxis=2))
    assert isinstance(ndc.extra_coords, ExtraCoords)
    ndc.extra_coords.add("time", 0, time_lut)

    cwcs = ndc.combined_wcs
    assert cwcs.world_n_dim == 3
    assert cwcs.pixel_n_dim == 2
    world = cwcs.pixel_to_world(0, 0)
    assert u.allclose(world[:2], (1, 1) * u.one)
    assert world[2] == Time("2011-01-01T00:00:00")


def test_slice_extra_1d(time_lut, wave_lut):
    ec = ExtraCoords()
    ec.add("time", 0, time_lut)
    ec.add("wavey", 1, wave_lut)

    sec = ec[:, 3:7]
    assert len(sec._lookup_tables) == 2

    assert u.allclose(sec['wavey'].wcs.pixel_to_world_values(list(range(4))),
                      ec['wavey'].wcs.pixel_to_world_values(list(range(3, 7))))
    assert u.allclose(sec['time'].wcs.pixel_to_world_values(list(range(4))),
                      ec['time'].wcs.pixel_to_world_values(list(range(4))))


def test_slice_extra_2d(time_lut, skycoord_2d_lut):
    ec = ExtraCoords()
    ec.add(("lat", "lon"), (0, 1), skycoord_2d_lut, mesh=False)
    ec.add("exposure_time", (0,), time_lut)

    sec = ec[1:5, 1:5]
    assert len(sec._lookup_tables) == 2

    assert u.allclose(sec['lat'].wcs.pixel_to_world_values(list(range(2)), list(range(2))),
                      ec['lat'].wcs.pixel_to_world_values(list(range(1, 3)), list(range(1, 3))))
    assert u.allclose(sec['lon'].wcs.pixel_to_world_values(list(range(2)), list(range(2))),
                      ec['lon'].wcs.pixel_to_world_values(list(range(1, 3)), list(range(1, 3))))

    assert u.allclose(sec['exposure_time'].wcs.pixel_to_world_values(list(range(0, 3))),
                      ec['exposure_time'].wcs.pixel_to_world_values(list(range(1, 4))))


def test_slice_drop_dimensions(time_lut, skycoord_2d_lut):
    ec = ExtraCoords()
    ec.add(("lat", "lon"), (0, 1), skycoord_2d_lut, mesh=False)
    ec.add("exposure_time", (0,), time_lut)

    sec = ec[0, :]
    assert len(sec._lookup_tables) == 1

    assert u.allclose(sec['lat'].wcs.pixel_to_world_values(list(range(2))),
                      ec['lat'].wcs.pixel_to_world_values([0, 0], list(range(2))))
    assert u.allclose(sec['lon'].wcs.pixel_to_world_values(list(range(2))),
                      ec['lon'].wcs.pixel_to_world_values([0, 0], list(range(2))))

    sec = ec[:, 0]
    assert len(sec._lookup_tables) == 2

    assert u.allclose(sec['lat'].wcs.pixel_to_world_values(list(range(2))),
                      ec['lat'].wcs.pixel_to_world_values(list(range(2)), [0, 0]))
    assert u.allclose(sec['lon'].wcs.pixel_to_world_values(list(range(2))),
                      ec['lon'].wcs.pixel_to_world_values(list(range(2)), [0, 0]))

    assert u.allclose(sec['exposure_time'].wcs.pixel_to_world_values(list(range(2)), list(range(2))),
                      ec['exposure_time'].wcs.pixel_to_world_values(list(range(2)), list(range(2))))


def test_slice_extra_twice(time_lut, wave_lut):
    ec = ExtraCoords()
    ec.add("time", 0, time_lut)
    ec.add("wavey", 1, wave_lut)

    sec = ec[1:, 0]
    assert len(sec._lookup_tables) == 1

    assert u.allclose(sec['time'].wcs.pixel_to_world_values(list(range(0, 2))),
                      ec['time'].wcs.pixel_to_world_values(list(range(1, 3))))

    sec = sec[1:, 0]
    assert len(sec._lookup_tables) == 1

    assert u.allclose(sec['time'].wcs.pixel_to_world_values(list(range(0, 2))),
                      ec['time'].wcs.pixel_to_world_values(list(range(2, 4))))


def test_slice_extra_1d_drop(time_lut, wave_lut):
    ec = ExtraCoords()
    ec.add("time", 0, time_lut)
    ec.add("wavey", 1, wave_lut)

    sec = ec[:, 3]
    assert len(sec._lookup_tables) == 1

    assert u.allclose(sec['time'].wcs.pixel_to_world_values(list(range(4))),
                      ec['time'].wcs.pixel_to_world_values(list(range(4))))

    dwd = sec.dropped_world_dimensions
    dwd.pop("world_axis_object_classes")
    assert dwd
    assert dwd["world_axis_units"] == ["nm"]


def test_slice_extra_1d_drop_alter_mapping_tuple_item(time_lut, wave_lut):
    ec = ExtraCoords()
    ec.add("time", 0, time_lut)
    ec.add("wavey", 1, wave_lut)

    sec = ec[0, :]
    assert len(sec._lookup_tables) == 1
    assert sec._lookup_tables[0][0] == (0,)
    assert u.allclose(sec['wavey'].wcs.pixel_to_world_values(list(range(10))),
                      ec['wavey'].wcs.pixel_to_world_values(list(range(10))))

    dwd = sec.dropped_world_dimensions
    dwd.pop("world_axis_object_classes")
    assert dwd
    assert dwd["world_axis_units"] == ["s"]


def test_slice_extra_1d_drop_alter_mapping_int_item(time_lut, wave_lut):
    ec = ExtraCoords()
    ec.add("time", 0, time_lut)
    ec.add("wavey", 1, wave_lut)

    sec = ec[0]
    assert len(sec._lookup_tables) == 1
    assert sec._lookup_tables[0][0] == (0,)
    assert u.allclose(sec['wavey'].wcs.pixel_to_world_values(list(range(10))),
                      ec['wavey'].wcs.pixel_to_world_values(list(range(10))))

    dwd = sec.dropped_world_dimensions
    dwd.pop("world_axis_object_classes")
    assert dwd
    assert dwd["world_axis_units"] == ["s"]


def test_dropped_dimension_reordering():
    data = np.ones((3, 4, 5))
    wcs_input_dict = {
        'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10, 'NAXIS1': 5,
        'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5, 'NAXIS2': 4,
        'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 3}
    input_wcs = WCS(wcs_input_dict)

    base_time = Time('2000-01-01', format='fits', scale='utc')
    timestamps = Time([base_time + TimeDelta(60 * i, format='sec') for i in range(data.shape[0])])

    my_cube = NDCube(data, input_wcs)
    my_cube.extra_coords.add('time', (0,), timestamps)

    # If the argument to extra_coords.add is array index then it should end up
    # in the first element of array_axis_physical_types
    assert "time" in my_cube.array_axis_physical_types[0]

    # When we slice out the dimension with the extra coord in it should go away.
    assert "time" not in my_cube[0].array_axis_physical_types[0]


def test_length1_extra_coord(wave_lut):
    # This test hits a bug that existed in gwcs less than 0.16.1
    pytest.importorskip("gwcs", minversion="0.16.1")
    ec = ExtraCoords()
    ec.add("wavey", 0, wave_lut)
    item = slice(1, 2)
    sec = ec[item]
    assert (sec.wcs.pixel_to_world(0) == wave_lut[item]).all()
    assert (sec.wcs.world_to_pixel(wave_lut[item])[0] == [0]).all()



import copy
import numbers
import textwrap

import astropy.units as u
import numpy as np

from ndcube import utils
from ndcube.visualization.descriptor import PlotterDescriptor

__all__ = ['NDCubeSequence']


class NDCubeSequenceBase:
    """
    Class representing a sequence of `~ndcube.NDCube`-like objects.

    The cubes are assumed to have the same dimensionality and axis physical types.

    Parameters
    ----------
    data_list : `list`
        List of `ndcube.NDCube`-like objects.

    meta : `dict` or None
        Meta data relevant to the sequence as a whole.

    common_axis: `int` or None
        The array axis of the cubes along which the cubes are ordered.
        For example, if the cubes are sequenced in chronological order and time is
        the 1st axis of each Cube, then common_axis should be set to 0.
        This enables the "cube_like" methods to be used, e.g.
        `ndcube.NDCubeSequence.index_as_cube` which slices the sequence as though it
        were a single cube concatenated along the common axis.
    """
    def __init__(self, data_list, meta=None, common_axis=None, **kwargs):
        self.data = data_list
        self.meta = meta
        if common_axis is not None:
            self._common_axis = int(common_axis)
        else:
            self._common_axis = common_axis

    @property
    def dimensions(self):
        """
        The length of each axis including the sequence axis.
        """
        return self._dimensions

    @property
    def _dimensions(self):
        dimensions = [len(self.data) * u.pix] + list(self.data[0].dimensions)
        if len(dimensions) > 1:
            # If there is a common axis, length of cube's along it may not
            # be the same. Therefore if the lengths are different,
            # represent them as a tuple of all the values, else as an int.
            if self._common_axis is not None:
                common_axis_lengths = [cube.data.shape[self._common_axis] for cube in self.data]
                if len(np.unique(common_axis_lengths)) != 1:
                    common_axis_dimensions = [cube.dimensions[self._common_axis]
                                              for cube in self.data]
                    dimensions[self._common_axis + 1] = u.Quantity(
                        common_axis_dimensions, unit=common_axis_dimensions[0].unit)
        return tuple(dimensions)

    @property
    def array_axis_physical_types(self):
        """
        The physical types associated with each array axis, including the sequence axis.
        """
        return [("meta.obs.sequence",)] + self.data[0].array_axis_physical_types

    @property
    def cube_like_dimensions(self):
        """
        The length of each array axis as if all cubes were concatenated along the common axis.
        """
        if not isinstance(self._common_axis, int):
            raise TypeError("Common axis must be set.")
        dimensions = list(self._dimensions)
        cube_like_dimensions = list(self._dimensions[1:])
        if dimensions[self._common_axis + 1].isscalar:
            cube_like_dimensions[self._common_axis] = u.Quantity(
                dimensions[0].value * dimensions[self._common_axis + 1].value, unit=u.pix)
        else:
            cube_like_dimensions[self._common_axis] = sum(dimensions[self._common_axis + 1])
        # Combine into single Quantity
        cube_like_dimensions = u.Quantity(cube_like_dimensions, unit=u.pix)
        return cube_like_dimensions

    @property
    def cube_like_array_axis_physical_types(self):
        """
        The physical types associated with each array axis, omitting the sequence axis.
        """
        if self._common_axis is None:
            raise ValueError("Common axis must be set.")
        return self.data[0].array_axis_physical_types

    def __getitem__(self, item):
        if isinstance(item, numbers.Integral):
            return self.data[item]
        # Create an empty sequence in which to place the sliced cubes.
        result = type(self)([], meta=self.meta, common_axis=self._common_axis)
        if isinstance(item, slice):
            result.data = self.data[item]
        else:
            if isinstance(item[0], numbers.Integral):
                result = self.data[item[0]][item[1:]]
            else:
                result.data = [cube[item[1:]] for cube in self.data[item[0]]]
            # Determine common axis after slicing.
            if self._common_axis is not None:
                drop_cube_axes = [isinstance(i, numbers.Integral) for i in item[1:]]
                if (len(drop_cube_axes) > self._common_axis and
                        drop_cube_axes[self._common_axis] is True):
                    result._common_axis = None
                else:
                    result._common_axis = \
                        self._common_axis - sum(drop_cube_axes[:self._common_axis])
        return result

    @property
    def index_as_cube(self):
        """
        Slice the NDCubesequence instance as a single cube concatenated along the common axis.

        Example
        -------
        >>> # Say we have three Cubes each cube has common_axis=0 is time and shape=(3,3,3)
        >>> data_list = [cubeA, cubeB, cubeC] # doctest: +SKIP
        >>> cs = NDCubeSequence(data_list, meta=None, common_axis=0) # doctest: +SKIP
        >>> # return zeroth time slice of cubeB in via normal NDCubeSequence indexing.
        >>> cs[1,:,0,:] # doctest: +SKIP
        >>> # Return same slice using this function
        >>> cs.index_as_cube[3:6, 0, :] # doctest: +SKIP
        """
        if self._common_axis is None:
            raise ValueError("common_axis cannot be None")
        return _IndexAsCubeSlicer(self)

    @property
    def common_axis_coords(self):
        """
        The coordinate values at each location along the common axis across all cubes.

        Only coordinates associated with the common axis in all cubes in the sequence
        are returned.  Coordinates from different cubes are concatenated along the
        common axis.  They thus represent the coordinate values at each location as
        if all cubes in the sequence were concatenated along the common axis.
        """
        common_axis = self._common_axis
        # Get coordinate objects associated with the common axis in all cubes.
        common_axis_names = set.intersection(*[set(cube.array_axis_physical_types[common_axis])
                                               for cube in self.data])
        common_coords = []
        mappings = []
        for i, cube in enumerate(self.data):
            cube_wcs = cube.combined_wcs
            common_coords.append(cube.axis_world_coords(common_axis, wcs=cube_wcs))
            mappings.append(utils.wcs.array_indices_for_world_objects(cube_wcs,
                                                                      axes=(common_axis,)))
        # For each coordinate, break up and then combine the coordinate objects across
        # the cubes into a list of coordinate objects that are length-1 and sequential
        # along the common axis.
        sequence_coords = []
        for coord_idx in range(len(common_coords[0])):
            exploded_coord = []
            for cube_idx in range(len(common_coords)):
                coord = common_coords[cube_idx][coord_idx]
                axis = np.where(np.array(mappings[cube_idx][coord_idx]) == common_axis)[0][0]
                item = [slice(None)] * len(coord.shape)
                for i in range(coord.shape[axis]):
                    item[axis] = i
                    exploded_coord.append(coord[tuple(item)])
            sequence_coords.append(exploded_coord)
        return sequence_coords

    @property
    def sequence_axis_coords(self):
        """
        Return the coordinate values along the sequence axis.

        These are compiled from the `~ndcube.GlobalCoords` objects attached to each
        `~ndcube.NDCube` where each cube represents a location along the sequence axis.
        Only coordinates that are common to all cubes are returned.
        """
        # Collect names of global coords common to all cubes.
        global_names = set.intersection(*[set(cube.global_coords.keys()) for cube in self.data])
        # For each coord, combine values from each cube's global coords property.
        return dict([(name, [cube.global_coords[name] for cube in self.data])
                     for name in global_names])

    def explode_along_axis(self, axis):
        """
        Separates slices of N-D cubes along a given cube axis into (N-1)D cubes.

        Parameters
        ----------
        axis : `int`
            The axis along which the data is to be changed.

        Returns
        -------
        `ndcube.NDCubeSequence`
            New sequence of (N-1)D cubes broken up along given axis.
        """
        # If axis is -ve then calculate the axis from the length of the dimensions of one cube.
        if axis < 0:
            axis = len(self.dimensions[1::]) + axis
        # To store the resultant cube
        result_cubes = []
        # All slices are initially initialised as slice(None, None, None)
        result_cubes_slice = [slice(None, None, None)] * len(self[0].data.shape)
        # the range of the axis that needs to be sliced
        range_of_axis = self[0].data.shape[axis]
        for ndcube in self.data:
            for index in range(range_of_axis):
                # setting the slice value to the index so that the slices are done correctly.
                result_cubes_slice[axis] = index
                # appending the sliced cubes in the result_cube list
                result_cubes.append(ndcube.__getitem__(tuple(result_cubes_slice)))
        # Determine common axis for new sequence.
        if self._common_axis is None or self._common_axis == axis:
            new_common_axis = None
        elif self._common_axis > axis:
            new_common_axis = self._common_axis - 1
        elif self._common_axis < axis:
            new_common_axis = self._common_axis
        # creating a new sequence with the result_cubes keeping the meta and common axis as axis
        return self._new_instance(result_cubes, common_axis=new_common_axis, meta=self.meta)

    def __str__(self):
        return (textwrap.dedent(f"""\
                NDCubeSequence
                --------------
                Dimensions:  {self.dimensions}
                Physical Types of Axes: {self.array_axis_physical_types}
                Common Cube Axis: {self._common_axis}"""))

    def __repr__(self):
        return f"{object.__repr__(self)}\n{str(self)}"

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    @classmethod
    def _new_instance(cls, data_list, meta=None, common_axis=None):
        """
        Instantiate a new instance of this class using given data.
        """
        return cls(data_list, meta=meta, common_axis=common_axis)


class NDCubeSequence(NDCubeSequenceBase):
    """
    Class representing a sequence of `~ndcube.NDCube`-like objects.

    The cubes are assumed to have the same dimensionality and axis physical types.

    Parameters
    ----------
    data_list : `list`
        List of `ndcube.NDCube`-like objects.

    meta : `dict` or None
        Meta data relevant to the sequence as a whole.

    common_axis: `int` or None
        The array axis of the cubes along which the cubes are ordered.
        For example, if the cubes are sequenced in chronological order and time is
        the 1st axis of each Cube, then common_axis should be set to 0.
        This enables the "cube_like" methods to be used, e.g.
        `ndcube.NDCubeSequence.index_as_cube` which slices the sequence as though it
        were a single cube concatenated along the common axis.
    """
    # We special case the default mpl plotter here so that we can only import
    # matplotlib when `.plotter` is accessed and raise an ImportError at the
    # last moment.
    plotter = PlotterDescriptor(default_type="mpl_sequence_plotter")

    def plot(self, *args, **kwargs):
        """
        A convenience function for the plotters default ``plot()`` method.

        Calling this method is the same as calling ``sequence.plotter.plot``, the
        behaviour of this method can change if the `NDCubeSequence.plotter` class is
        set to a different ``Plotter`` class.

        """
        if self.plotter is None:
            raise NotImplementedError(
                "This NDCubeSequence object does not have a .plotter defined so "
                "no default plotting functionality is available.")

        return self.plotter.plot(*args, **kwargs)

    def plot_as_cube(self, *args, **kwargs):
        raise NotImplementedError(
            "NDCubeSequence plot_as_cube is no longer supported.\n"
            "To learn why or to tell us why it should be re-instated, "
            "read and comment on issue #315:\n\nhttps://github.com/sunpy/ndcube/issues/315\n\n"
            "To see a introductory guide on how to make your own NDCubeSequence plots, "
            "see the docs:\n\n"
            "https://docs.sunpy.org/projects/ndcube/en/stable/ndcubesequence.html#plotting")


"""
Cube Sequence Helpers
"""


class _IndexAsCubeSlicer:
    """
    Helper class to make slicing in index_as_cube sliceable/indexable like a
    numpy array.

    Parameters
    ----------
    seq : `ndcube.NDCubeSequence`
        Object of NDCubeSequence.
    """

    def __init__(self, seq):
        self.seq = seq

    def __getitem__(self, item):
        common_axis = self.seq._common_axis
        common_axis_lengths = [int(cube.dimensions[common_axis].value) for cube in self.seq.data]
        n_cube_dims = len(self.seq.cube_like_dimensions)
        n_uncommon_cube_dims = n_cube_dims - 1
        # If item is iint or slice, turn into a tuple, filling in items
        # for unincluded axes with slice(None). This ensures it is
        # treated the same as tuple items.
        if isinstance(item, (numbers.Integral, slice)):
            item = [item] + [slice(None)] * n_uncommon_cube_dims
        else:
            # Item must therefore be tuple. Ensure it has an entry for each axis.
            item = list(item) + [slice(None)] * (n_cube_dims - len(item))
        # If common axis item is slice(None), result is trivial as common_axis is not changed.
        if item[common_axis] == slice(None):
            # Create item for slicing through the default API and slice.
            return self.seq[tuple([slice(None)] + item)]
        if isinstance(item[common_axis], numbers.Integral):
            # If common_axis item is an int or return an NDCube with dimensionality of N-1
            sequence_index, common_axis_index = \
                utils.sequence.cube_like_index_to_sequence_and_common_axis_indices(
                    item[common_axis], common_axis, common_axis_lengths)
            # Insert index for common axis in item for slicing the NDCube.
            cube_item = copy.deepcopy(item)
            cube_item[common_axis] = common_axis_index
            return self.seq.data[sequence_index][tuple(cube_item)]
        else:
            # item can now only be a tuple whose common axis item is a non-None slice object.
            # Convert item into iterable of SequenceItems and slice each cube appropriately.
            # item for common_axis must always be a slice for every cube,
            # even if it is only a length-1 slice.
            # Thus NDCubeSequence.index_as_cube can only slice away common axis if
            # item is int or item's first item is an int.
            # i.e. NDCubeSequence.index_as_cube cannot cause common_axis to become None
            # since in all cases where the common_axis is sliced away involve an NDCube
            # is returned, not an NDCubeSequence.
            # common_axis of returned sequence must be altered if axes in front of it
            # are sliced away.
            sequence_items = utils.sequence.cube_like_tuple_item_to_sequence_items(
                item, common_axis, common_axis_lengths, n_cube_dims)
            # Work out new common axis value if axes in front of it are sliced away.
            new_common_axis = common_axis - sum([isinstance(i, numbers.Integral)
                                                 for i in item[:common_axis]])
            # Copy sequence and alter the data and common axis.
            result = type(self.seq)([], meta=self.seq.meta, common_axis=new_common_axis)
            result.data = [self.seq.data[sequence_item.sequence_index][sequence_item.cube_item]
                           for sequence_item in sequence_items]
            return result



import numpy as np
from astropy.wcs.wcsapi.wrappers.base import BaseWCSWrapper

__all__ = ['ResampledLowLevelWCS']


class ResampledLowLevelWCS(BaseWCSWrapper):
    """
    A wrapper for a low-level WCS object that has down- or
    up-sampled pixel axes.

    Parameters
    ----------
    wcs : `~astropy.wcs.wcsapi.BaseLowLevelWCS`
        The original WCS for which to reorder axes

    factor : `int` or `float` or iterable of the same
        The factor by which to increase the pixel size for each pixel
        axis. If a scalar, the same factor is used for all axes.

    offset: `int` or `float` or iterable of the same
        The location on the underlying pixel grid which corresponds
        to zero on the top level pixel grid. If a scalar, the grid will be
        shifted by the same amount in all dimensions.
    """
    def __init__(self, wcs, factor, offset=0):
        self._wcs = wcs
        if np.isscalar(factor):
            factor = [factor] * self.pixel_n_dim
        self._factor = np.array(factor)
        if len(self._factor) !=  self.pixel_n_dim:
            raise ValueError(f"Length of factor must equal number of dimensions {self.pixel_n_dim}.")
        if np.isscalar(offset):
            offset = [offset] * self.pixel_n_dim
        self._offset = np.array(offset)
        if len(self._offset) != self.pixel_n_dim:
            raise ValueError(f"Length of offset must equal number of dimensions {self.pixel_n_dim}.")

    def _top_to_underlying_pixels(self, top_pixels):
        # Convert user-facing pixel indices to the pixel grid of underlying WCS.
        factor = self._pad_dims(self._factor, top_pixels.ndim)
        offset = self._pad_dims(self._offset, top_pixels.ndim)
        return top_pixels * factor + offset

    def _underlying_to_top_pixels(self, underlying_pixels):
        # Convert pixel indices of underlying pixel grid to user-facing grid.
        factor = self._pad_dims(self._factor, underlying_pixels.ndim)
        offset = self._pad_dims(self._offset, underlying_pixels.ndim)
        return (underlying_pixels - offset) / factor

    def _pad_dims(self, arr, ndim):
        # Pad array with trailing degenerate dimensions.
        # This make scaling with pixel arrays easier.
        shape = np.ones(ndim, dtype=int)
        shape[0] = len(arr)
        return arr.reshape(tuple(shape))

    def pixel_to_world_values(self, *pixel_arrays):
        underlying_pixel_arrays = self._top_to_underlying_pixels(np.asarray(pixel_arrays))
        return self._wcs.pixel_to_world_values(*underlying_pixel_arrays)

    def world_to_pixel_values(self, *world_arrays):
        underlying_pixel_arrays = self._wcs.world_to_pixel_values(*world_arrays)
        top_pixel_arrays = self._underlying_to_top_pixels(np.asarray(underlying_pixel_arrays))
        return tuple(array for array in top_pixel_arrays)

    @property
    def pixel_shape(self):
        # Return pixel shape of resampled grid.
        # Where shape is an integer, return an int type as its required for some uses.
        if self._wcs.pixel_shape is None:
            return self._wcs.pixel_shape
        underlying_shape = np.asarray(self._wcs.pixel_shape)
        int_elements = np.isclose(np.mod(underlying_shape, self._factor), 0,
                                  atol=np.finfo(float).resolution)
        pixel_shape = underlying_shape / self._factor
        return tuple(int(np.rint(i)) if is_int else i
                     for i, is_int in zip(pixel_shape, int_elements))

    @property
    def pixel_bounds(self):
        if self._wcs.pixel_bounds is None:
            return self._wcs.pixel_bounds
        top_level_bounds = self._underlying_to_top_pixels(np.asarray(self._wcs.pixel_bounds))
        return [tuple(bounds) for bounds in top_level_bounds]



import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.coordinates.spectral_coordinate import SpectralCoord
from astropy.wcs.wcsapi.high_level_api import HighLevelWCSMixin
from astropy.wcs.wcsapi.low_level_api import BaseLowLevelWCS

from ndcube.global_coords import GlobalCoords
from ndcube.ndcube import NDCube


@pytest.fixture
def gc():
    return GlobalCoords()


@pytest.fixture
def gc_coords(gc):
    coord1 = 1 * u.m
    coord2 = 2 * u.s
    gc.add('name1', 'custom:physical_type1', coord1)
    gc.add('name2', 'custom:physical_type2', coord2)
    return gc


def test_add(gc):
    coord1 = 1 * u.m
    coord2 = 2 * u.s
    gc.add('name1', 'custom:physical_type1', coord1)
    gc.add('name2', 'custom:physical_type2', coord2)
    assert gc.keys() == {'name1', 'name2'}
    assert gc.physical_types == dict((('name1', 'custom:physical_type1'), ('name2', 'custom:physical_type2')))


def test_remove(gc_coords):
    gc_coords.remove('name2')
    assert len(gc_coords) == 1
    assert gc_coords.keys() == {'name1'}
    assert gc_coords.physical_types == {'name1': 'custom:physical_type1'}


def test_overwrite(gc_coords):
    with pytest.raises(ValueError):
        coord2 = 2 * u.s
        gc_coords.add('name1', 'custom:physical_type2', coord2)


def test_iterating(gc_coords):
    for i, gc_item in enumerate(gc_coords):
        if i == 0:
            assert gc_item == 'name1'
        if i == 1:
            assert gc_item == 'name2'


def test_slicing(gc_coords):
    assert u.allclose(gc_coords['name1'], u.Quantity(1., u.m))


def test_physical_types(gc_coords):
    assert gc_coords.physical_types == dict((('name1', 'custom:physical_type1'), ('name2', 'custom:physical_type2')))


def test_len(gc_coords):
    assert len(gc_coords) == 2


def test_keys(gc_coords):
    assert gc_coords.keys() == {'name1', 'name2'}


def test_values(gc_coords):
    for value, expected in zip(gc_coords.values(), (1 * u.m, 2 * u.s)):
        assert u.allclose(value, expected)


def test_items(gc_coords):
    assert gc_coords.items() == {('name1', 1 * u.m), ('name2', 2 * u.s)}


def test_filter(gc_coords):
    filtered = gc_coords.filter_by_physical_type('custom:physical_type1')
    assert isinstance(filtered, GlobalCoords)
    assert len(filtered) == 1
    assert 'name1' in filtered
    assert u.allclose(filtered['name1'], 1 * u.m)
    assert filtered.physical_types == {'name1': 'custom:physical_type1'}


def test_dropped_to_global(ndcube_4d_ln_l_t_lt):
    ndcube_4d_ln_l_t_lt.wcs.wcs.cname = ['lat', 'time', 'wavelength', 'lon']
    sub = ndcube_4d_ln_l_t_lt[0, 0, :, 0]
    gc = sub.global_coords
    assert len(gc) == 2

    assert isinstance(gc["helioprojective"], SkyCoord)
    assert isinstance(gc["wavelength"], SpectralCoord)


def test_dropped_to_global_ec(ndcube_4d_ln_lt_l_t):
    ndcube_4d_ln_lt_l_t.extra_coords.add("test1", 0, np.arange(ndcube_4d_ln_lt_l_t.data.shape[0]) * u.m)
    sub = ndcube_4d_ln_lt_l_t[0, 0, :, :]
    gc = sub.global_coords
    assert len(gc) == 2

    assert isinstance(gc["helioprojective"], SkyCoord)
    assert isinstance(gc["test1"], u.Quantity)


def test_dropped_to_global_ec_gwcs_fail(ndcube_4d_extra_coords):
    sub = ndcube_4d_extra_coords[0, 0, :, :]
    gc = sub.global_coords

    pytest.importorskip("gwcs", minversion="0.16.2a1.dev23+g14a8c0b")

    # Due to https://github.com/spacetelescope/gwcs/issues/358 one of the two
    # extra coords overwrites the other here.
    assert len(gc) == 3

    assert isinstance(gc["helioprojective"], SkyCoord)
    assert isinstance(gc["time"], u.Quantity)
    assert isinstance(gc["hello"], u.Quantity)


class SerializedWCS(BaseLowLevelWCS, HighLevelWCSMixin):
    """
    WCS with serialized classes
    """

    def pixel_to_world_values(self, *pixel_arrays):
        return [np.asarray(pix) * 2 for pix in pixel_arrays]

    def world_to_pixel_values(self, *world_arrays):
        return [np.asarray(world) / 2 for world in world_arrays]

    @property
    def serialized_classes(self):
        return True

    @property
    def pixel_n_dim(self):
        return 2

    @property
    def world_n_dim(self):
        return 2

    @property
    def axis_correlation_matrix(self):
        return np.identity(2)

    @property
    def world_axis_physical_types(self):
        return ['pos.eq.ra', 'pos.eq.dec']

    @property
    def world_axis_units(self):
        return ['deg', 'deg']

    @property
    def world_axis_object_components(self):
        return [('test', 0, 'value'), ('test2', 0, 'value')]

    @property
    def world_axis_object_classes(self):
        return {'test': ('astropy.units.Quantity', (),
                         {'unit': ('astropy.units.Unit', ('deg',), {})}),
                'test2': ('astropy.units.Quantity', (),
                          {'unit': ('astropy.units.Unit', ('deg',), {})})}


def test_serialized_classes():
    ndc = NDCube(np.zeros((10, 10)), wcs=SerializedWCS())
    assert u.allclose(ndc[0, :].global_coords["pos.eq.dec"], 0*u.deg)


class MultiCoord(list):
    def __init__(self, *args):
        super().__init__(args)

    @property
    def value(self):
        return tuple(x.value() for x in self)


class MultiCoordWCS(BaseLowLevelWCS, HighLevelWCSMixin):
    """
    WCS with serialized classes
    """

    def pixel_to_world_values(self, *pixel_arrays):
        return [np.asarray(pix) * 2 for pix in pixel_arrays]

    def world_to_pixel_values(self, *world_arrays):
        return [np.asarray(world) / 2 for world in world_arrays]

    @property
    def pixel_n_dim(self):
        return 3

    @property
    def world_n_dim(self):
        return 3

    @property
    def axis_correlation_matrix(self):
        return np.array([[1, 1, 0],
                         [1, 1, 0],
                         [0, 0, 1]])

    @property
    def world_axis_physical_types(self):
        return ['pos.eq.ra', 'pos.eq.dec', 'pos.distance.x']

    @property
    def world_axis_units(self):
        return ['deg', 'deg', 'm']

    @property
    def world_axis_object_components(self):
        return [('test', 0, 'value'),
                ('test', 1, 'value'),
                ('distance', 0, 'value')]

    @property
    def world_axis_object_classes(self):
        return {'test': (MultiCoord, (), {}),
                'distance': (u.Quantity, (), {'unit': 'm'})}


def test_non_skycoord_multi_object():
    ndc = NDCube(np.zeros((10, 20, 30)), wcs=MultiCoordWCS())
    gc = ndc[:, 0, 0].global_coords
    assert len(gc._all_coords) == 1
    assert gc["pos.eq.ra"] == gc["pos.eq.dec"] == gc[("pos.eq.ra", "pos.eq.dec")] == [0, 0]
    assert not set(gc.keys()).difference({("pos.eq.ra", "pos.eq.dec")})



import unittest

import astropy.units as u
import numpy as np
import pytest
from astropy.time import Time, TimeDelta

from ndcube import NDCube, NDCubeSequence


def derive_sliced_cube_dims(orig_cube_dims, tuple_item):
    expected_cube_dims = list(orig_cube_dims)
    len_cube_item = len(tuple_item) - 1
    if len_cube_item > 0:
        cube_item = tuple_item[1:]
        for i, s in zip(np.arange(len_cube_item)[::-1], cube_item[::-1]):
            if isinstance(s, int):
                del expected_cube_dims[i]
            else:
                expected_cube_dims[i] = float(s.stop - s.start) * u.pix
    expected_cube_dims *= u.pix
    return expected_cube_dims


@pytest.mark.parametrize("ndc, item",
                         (
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:1], ),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:1, 0:2]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:1, 1]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[1:3, 1, 0:2])
                         ),
                         indirect=('ndc',))
def test_slice_sequence_axis(ndc, item):
    # Calculate expected dimensions of cubes with sequence after slicing.
    tuple_item = item if isinstance(item, tuple) else (item,)
    expected_cube0_dims = derive_sliced_cube_dims(ndc.data[tuple_item[0]][0].dimensions, tuple_item)
    # Assert output is as expected.
    sliced_sequence = ndc[item]
    assert isinstance(sliced_sequence, NDCubeSequence)
    assert int(sliced_sequence.dimensions[0].value) == tuple_item[0].stop - tuple_item[0].start
    assert (sliced_sequence[0].dimensions == expected_cube0_dims).all()


@pytest.mark.parametrize("ndc, item",
                         (
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[1, 0:1]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[2, 1]),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[3, 1, 0:2])
                         ),
                         indirect=("ndc",))
def test_extract_ndcube(ndc, item):
    cube = ndc[item]
    tuple_item = item if isinstance(item, tuple) else (item,)
    expected_cube_dims = derive_sliced_cube_dims(ndc.data[tuple_item[0]].dimensions, tuple_item)
    assert isinstance(cube, NDCube)
    assert (cube.dimensions == expected_cube_dims).all()


@pytest.mark.parametrize("ndc, item, expected_common_axis",
                         (
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, 0], 0),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, 0:1, 0:2], 1),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, :, :, 1], 1),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[:, :, 0], None)
                         ),
                         indirect=("ndc",))
def test_slice_common_axis(ndc, item, expected_common_axis):
    sliced_sequence = ndc[item]
    assert sliced_sequence._common_axis == expected_common_axis


@pytest.mark.parametrize("ndc, item, expected_dimensions",
                         (
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[1:7], (3 * u.pix,
                                                                             2 * u.pix,
                                                                             [2., 3., 1.] * u.pix,
                                                                             4 * u.pix)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[1:7, 0], (3 * u.pix,
                                                                                [2., 3., 1.] * u.pix,
                                                                                4 * u.pix)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[2:4], (2 * u.pix,
                                                                             2 * u.pix,
                                                                             1 * u.pix,
                                                                             4 * u.pix)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:6], (2 * u.pix,
                                                                             2 * u.pix,
                                                                             3 * u.pix,
                                                                             4 * u.pix)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0:6, :, 0], (3 * u.pix,
                                                                                   2 * u.pix,
                                                                                   4 * u.pix)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", np.s_[0::, 0, 0], (4 * u.pix,
                                                                                   2 * u.pix))
                         ),
                         indirect=("ndc",))
def test_index_as_cube(ndc, item, expected_dimensions):
    sliced_sequence = ndc.index_as_cube[item]
    sliced_dims = sliced_sequence.dimensions
    for dim, expected_dim in zip(sliced_dims, expected_dimensions):
        (dim == expected_dim).all()


@pytest.mark.parametrize("ndc, axis, expected_dimensions",
                         (
                             ("ndcubesequence_4c_ln_lt_l", 0, (8 * u.pix,
                                                               3 * u.pix,
                                                               4 * u.pix)),
                             ("ndcubesequence_4c_ln_lt_l_cax1", 1, (12 * u.pix,
                                                                    2 * u.pix,
                                                                    4 * u.pix))
                         ),
                         indirect=("ndc",))
def test_explode_along_axis_common_axis_None(ndc, axis, expected_dimensions):
    exploded_sequence = ndc.explode_along_axis(axis)
    assert exploded_sequence.dimensions == expected_dimensions
    assert exploded_sequence._common_axis is None


@pytest.mark.parametrize("ndc", (('ndcubesequence_4c_ln_lt_l_cax1',)), indirect=("ndc",))
def test_explode_along_axis_common_axis_same(ndc):
    exploded_sequence = ndc.explode_along_axis(2)
    assert exploded_sequence.dimensions == (16*u.pix, 2*u.pix, 3*u.pix)
    assert exploded_sequence._common_axis == ndc._common_axis


@pytest.mark.parametrize("ndc", (('ndcubesequence_4c_ln_lt_l_cax1',)), indirect=("ndc",))
def test_explode_along_axis_common_axis_changed(ndc):
    exploded_sequence = ndc.explode_along_axis(0)
    assert exploded_sequence.dimensions == (8*u.pix, 3*u.pix, 4*u.pix)
    assert exploded_sequence._common_axis == ndc._common_axis - 1


@pytest.mark.parametrize("ndc, expected_dimensions",
                         (
                             ("ndcubesequence_4c_ln_lt_l_cax1", (4 * u.pix,
                                                                 2. * u.pix,
                                                                 3. * u.pix,
                                                                 4. * u.pix)),
                         ),
                         indirect=("ndc",))
def test_dimensions(ndc, expected_dimensions):
    unit_tester = unittest.TestCase()
    unit_tester.assertEqual(ndc.dimensions, expected_dimensions)


@pytest.mark.parametrize("ndc, expected_dimensions",
                         (
                             ("ndcubesequence_4c_ln_lt_l_cax1", [2., 12, 4] * u.pix),
                         ),
                         indirect=("ndc",))
def test_cube_like_dimensions(ndc, expected_dimensions):
    assert (ndc.cube_like_dimensions == expected_dimensions).all()


@pytest.mark.parametrize("ndc", (("ndcubesequence_4c_ln_lt_l",)), indirect=("ndc",))
def test_cube_like_dimensions_error(ndc):
    with pytest.raises(TypeError):
        ndc.cube_like_dimensions


@pytest.mark.parametrize("ndc", (("ndcubesequence_3c_l_ln_lt_cax1",)), indirect=("ndc",))
def test_common_axis_coords(ndc):
    # Construct expected skycoord
    common_coords = [cube.axis_world_coords('lon') for cube in ndc]
    expected_skycoords = []
    for cube_coords in common_coords:
        expected_skycoords += [cube_coords[0][i] for i in range(len(cube_coords[0]))]
    # Construct expected Times
    base_time = Time('2000-01-01', format='fits', scale='utc')
    expected_times = [base_time + TimeDelta(60*i, format='sec') for i in range(15)]
    # Run test function.
    output = ndc.common_axis_coords
    #  Check right number of coords returned.
    assert len(output) == 2
    output_skycoords, output_times = output
    # Check SkyCoords are equal.
    for output_coord, expected_coord in zip(output_skycoords, expected_skycoords):
        assert all(output_coord == expected_coord)
    # Check times are equal
    for output_time, expected_time in zip(output_times, expected_times):
        td = output_time - expected_time
        assert u.allclose(td.to(u.s), 0*u.s, atol=1e-10*u.s)


@pytest.mark.parametrize("ndc", (("ndcubesequence_3c_l_ln_lt_cax1",)), indirect=("ndc",))
def test_sequence_axis_coords(ndc):
    expected = {'distance': [1*u.m, 2*u.m, 3*u.m]}
    output = ndc.sequence_axis_coords
    assert output == expected



# import pytest
# import datetime
# import copy

# import numpy as np
# import astropy.units as u
# import matplotlib

# from ndcube import NDCube, NDCubeSequence
# from ndcube.utils.wcs import WCS
# import ndcube.mixins.sequence_plotting

# # sample data for tests
# # TODO: use a fixture reading from a test file. file TBD.
# data = np.array([[[1, 2, 3, 4], [2, 4, 5, 3], [0, -1, 2, 3]],
#                  [[2, 4, 5, 1], [10, 5, 2, 2], [10, 3, 3, 0]]])

# data2 = np.array([[[11, 22, 33, 44], [22, 44, 55, 33], [0, -1, 22, 33]],
#                   [[22, 44, 55, 11], [10, 55, 22, 22], [10, 33, 33, 0]]])

# ht = {'CTYPE3': 'HPLT-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.5, 'CRPIX3': 0, 'CRVAL3': 0, 'NAXIS3': 2,
#       'CTYPE2': 'WAVE    ', 'CUNIT2': 'Angstrom', 'CDELT2': 0.2, 'CRPIX2': 0, 'CRVAL2': 0,
#       'NAXIS2': 3,
#       'CTYPE1': 'TIME    ', 'CUNIT1': 'min', 'CDELT1': 0.4, 'CRPIX1': 0, 'CRVAL1': 0, 'NAXIS1': 4}
# wt = WCS(header=ht, naxis=3)

# hm = {
#     'CTYPE1': 'WAVE    ', 'CUNIT1': 'Angstrom', 'CDELT1': 0.2, 'CRPIX1': 0, 'CRVAL1': 10,
#     'NAXIS1': 4,
#     'CTYPE2': 'HPLT-TAN', 'CUNIT2': 'deg', 'CDELT2': 0.5, 'CRPIX2': 2, 'CRVAL2': 0.5, 'NAXIS2': 3,
#     'CTYPE3': 'HPLN-TAN', 'CUNIT3': 'deg', 'CDELT3': 0.4, 'CRPIX3': 2, 'CRVAL3': 1, 'NAXIS3': 2}
# wm = WCS(header=hm, naxis=3)


# cube1 = NDCube(
#     data, wt, missing_axes=[False, False, False, True],
#     extra_coords=[
#         ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
#         ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(0, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# cube1_with_unit = NDCube(
#     data, wt, missing_axes=[False, False, False, True],
#     unit=u.km,
#     extra_coords=[
#         ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
#         ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(0, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# cube1_with_mask = NDCube(
#     data, wt, missing_axes=[False, False, False, True],
#     mask=np.zeros_like(data, dtype=bool),
#     extra_coords=[
#         ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
#         ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(0, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# cube1_with_uncertainty = NDCube(
#     data, wt, missing_axes=[False, False, False, True],
#     uncertainty=np.sqrt(data),
#     extra_coords=[
#         ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
#         ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(0, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# cube1_with_unit_and_uncertainty = NDCube(
#     data, wt, missing_axes=[False, False, False, True],
#     unit=u.km, uncertainty=np.sqrt(data),
#     extra_coords=[
#         ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
#         ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(0, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# cube3 = NDCube(
#     data2, wt, missing_axes=[False, False, False, True],
#     extra_coords=[
#         ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
#          cube1.extra_coords['pix']['value'][-1]),
#         ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(2, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

# cube3_with_unit = NDCube(
#     data2, wt, missing_axes=[False, False, False, True],
#     unit=u.m,
#     extra_coords=[
#         ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
#          cube1.extra_coords['pix']['value'][-1]),
#         ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(2, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

# cube3_with_mask = NDCube(
#     data2, wt, missing_axes=[False, False, False, True],
#     mask=np.zeros_like(data2, dtype=bool),
#     extra_coords=[
#         ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
#          cube1.extra_coords['pix']['value'][-1]),
#         ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(2, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

# cube3_with_uncertainty = NDCube(
#     data2, wt, missing_axes=[False, False, False, True],
#     uncertainty=np.sqrt(data2),
#     extra_coords=[
#         ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
#          cube1.extra_coords['pix']['value'][-1]),
#         ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(2, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

# cube3_with_unit_and_uncertainty = NDCube(
#     data2, wt, missing_axes=[False, False, False, True],
#     unit=u.m, uncertainty=np.sqrt(data2),
#     extra_coords=[
#         ('pix', 0, u.Quantity(np.arange(1, data2.shape[0]+1), unit=u.pix) +
#          cube1.extra_coords['pix']['value'][-1]),
#         ('hi', 1, u.Quantity(range(data2.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(2, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 2))])

# cubem1 = NDCube(
#     data, wm,
#     extra_coords=[
#         ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
#         ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(0, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# cubem3 = NDCube(
#     data2, wm,
#     extra_coords=[
#         ('pix', 0, u.Quantity(range(data.shape[0]), unit=u.pix)),
#         ('hi', 1, u.Quantity(range(data.shape[1]), unit=u.s)),
#         ('distance', None, u.Quantity(0, unit=u.cm)),
#         ('time', None, datetime.datetime(2000, 1, 1, 0, 0))])

# # Define some test NDCubeSequences.
# common_axis = 0
# seq = NDCubeSequence(data_list=[cube1, cube3, cube1, cube3], common_axis=common_axis)

# seq_no_common_axis = NDCubeSequence(data_list=[cube1, cube3, cube1, cube3])

# seq_with_units = NDCubeSequence(
#     data_list=[cube1_with_unit, cube3_with_unit, cube1_with_unit, cube3_with_unit],
#     common_axis=common_axis)

# seq_with_masks = NDCubeSequence(
#     data_list=[cube1_with_mask, cube3_with_mask, cube1_with_mask, cube3_with_mask],
#     common_axis=common_axis)

# seq_with_unit0 = NDCubeSequence(data_list=[cube1_with_unit, cube3,
#                                            cube1_with_unit, cube3], common_axis=common_axis)

# seq_with_mask0 = NDCubeSequence(data_list=[cube1_with_mask, cube3,
#                                            cube1_with_mask, cube3], common_axis=common_axis)

# seq_with_uncertainty = NDCubeSequence(data_list=[cube1_with_uncertainty, cube3_with_uncertainty,
#                                                  cube1_with_uncertainty, cube3_with_uncertainty],
#                                       common_axis=common_axis)

# seq_with_some_uncertainty = NDCubeSequence(
#     data_list=[cube1_with_uncertainty, cube3, cube1, cube3_with_uncertainty],
#     common_axis=common_axis)

# seq_with_units_and_uncertainty = NDCubeSequence(
#     data_list=[cube1_with_unit_and_uncertainty, cube3_with_unit_and_uncertainty,
#                cube1_with_unit_and_uncertainty, cube3_with_unit_and_uncertainty],
#     common_axis=common_axis)

# seq_with_units_and_some_uncertainty = NDCubeSequence(
#     data_list=[cube1_with_unit_and_uncertainty, cube3_with_unit,
#                cube1_with_unit, cube3_with_unit_and_uncertainty],
#     common_axis=common_axis)

# seq_with_some_masks = NDCubeSequence(data_list=[cube1_with_mask, cube3, cube1, cube3_with_mask],
#                                      common_axis=common_axis)

# seqm = NDCubeSequence(data_list=[cubem1, cubem3, cubem1, cubem3], common_axis=common_axis)

# # Derive some expected data arrays in plot objects.
# seq_data_stack = np.stack([cube.data for cube in seq_with_masks.data])
# seq_mask_stack = np.stack([cube.mask for cube in seq_with_masks.data])

# seq_stack = np.ma.masked_array(seq_data_stack, seq_mask_stack)
# seq_stack_km = np.ma.masked_array(
#     np.stack([(cube.data * cube.unit).to(u.km).value for cube in seq_with_units.data]),
#     seq_mask_stack)

# seq_data_concat = np.concatenate([cube.data for cube in seq_with_masks.data], axis=common_axis)
# seq_mask_concat = np.concatenate([cube.mask for cube in seq_with_masks.data], axis=common_axis)

# seq_concat = np.ma.masked_array(seq_data_concat, seq_mask_concat)
# seq_concat_km = np.ma.masked_array(
#     np.concatenate([(cube.data * cube.unit).to(u.km).value
#                     for cube in seq_with_units.data], axis=common_axis),
#     seq_mask_concat)

# # Derive expected axis_ranges for non-cube-like cases.
# x_axis_coords3 = np.array([0.4, 0.8, 1.2, 1.6]).reshape((1, 1, 4))
# new_x_axis_coords3_shape = u.Quantity(seq.dimensions, unit=u.pix).value.astype(int)
# new_x_axis_coords3_shape[-1] = 1
# none_axis_ranges_axis3 = [np.arange(0, len(seq.data)+1),
#                           np.array([0., 1., 2.]), np.arange(0, 4),
#                           np.tile(np.array(x_axis_coords3), new_x_axis_coords3_shape)]
# none_axis_ranges_axis0 = [np.arange(len(seq.data)),
#                           np.array([0., 1., 2.]), np.arange(0, 4),
#                           np.arange(0, int(seq.dimensions[-1].value)+1)]
# distance0_none_axis_ranges_axis0 = [seq.sequence_axis_extra_coords["distance"].value,
#                                     np.array([0., 1., 2.]), np.arange(0, 4),
#                                     np.arange(0, int(seq.dimensions[-1].value)+1)]
# distance0_none_axis_ranges_axis0_mm = [seq.sequence_axis_extra_coords["distance"].to("mm").value,
#                                        np.array([0., 1., 2.]), np.arange(0, 4),
#                                        np.arange(0, int(seq.dimensions[-1].value)+1)]
# userrangequantity_none_axis_ranges_axis0 = [
#     np.arange(int(seq.dimensions[0].value)), np.array([0., 1., 2.]), np.arange(0, 4),
#     np.arange(0, int(seq.dimensions[-1].value)+1)]

# userrangequantity_none_axis_ranges_axis0_1e7 = [
#     (np.arange(int(seq.dimensions[0].value)) * u.J).to(u.erg).value, np.array([0., 1., 2.]),
#     np.arange(0, 4), np.arange(0, int(seq.dimensions[-1].value)+1)]

# hi2_none_axis_ranges_axis2 = [
#     np.arange(0, len(seq.data)+1), np.array([0., 1., 2.]),
#     np.arange(int(seq.dimensions[2].value)), np.arange(0, int(seq.dimensions[-1].value)+1)]

# x_axis_coords1 = np.zeros(tuple([int(s.value) for s in seq.dimensions]))
# x_axis_coords1[0, 1] = 1.
# x_axis_coords1[1, 0] = 2.
# x_axis_coords1[1, 1] = 3.
# x_axis_coords1[2, 1] = 1.
# x_axis_coords1[3, 0] = 2.
# x_axis_coords1[3, 1] = 3.
# pix1_none_axis_ranges_axis1 = [
#     np.arange(0, len(seq.data)+1), x_axis_coords1, np.arange(0, 4),
#     np.arange(0, int(seq.dimensions[-1].value)+1)]

# # Derive expected extents
# seq_axis1_lim_deg = [0.49998731, 0.99989848]
# seq_axis1_lim_arcsec = [(axis1_xlim*u.deg).to(u.arcsec).value for axis1_xlim in seq_axis1_lim_deg]
# seq_axis2_lim_m = [seq[:, :, :, 0].data[0].axis_world_coords()[-1][0].value,
#                    seq[:, :, :, 0].data[0].axis_world_coords()[-1][-1].value]

# # Derive expected axis_ranges for cube-like cases.
# cube_like_new_x_axis_coords2_shape = u.Quantity(
#     seq.cube_like_dimensions, unit=u.pix).value.astype(int)
# cube_like_new_x_axis_coords2_shape[-1] = 1
# cubelike_none_axis_ranges_axis2 = [
#     np.arange(0, int(seq.cube_like_dimensions[0].value)+1), np.arange(0, 4),
#     np.tile(x_axis_coords3, cube_like_new_x_axis_coords2_shape)]

# cubelike_none_axis_ranges_axis2_s = copy.deepcopy(cubelike_none_axis_ranges_axis2)
# cubelike_none_axis_ranges_axis2_s[2] = cubelike_none_axis_ranges_axis2_s[2] * 60.

# cubelike_none_axis_ranges_axis0 = [[0, 8], np.arange(0, 4),
#                                    np.arange(0, int(seq.cube_like_dimensions[-1].value)+1)]


# @pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
#     (seq[:, 0, 0, 0], {},
#      (np.arange(len(seq.data)), np.array([1, 11,  1, 11]),
#       "meta.obs.sequence [None]", "Data [None]", (0, len(seq[:, 0, 0, 0].data)-1),
#       (min([cube.data.min() for cube in seq[:, 0, 0, 0].data]),
#        max([cube.data.max() for cube in seq[:, 0, 0, 0].data])))),

#     (seq_with_units[:, 0, 0, 0], {},
#      (np.arange(len(seq_with_units.data)), np.array([1, 0.011,  1, 0.011]),
#       "meta.obs.sequence [None]", "Data [km]", (0, len(seq_with_units[:, 0, 0, 0].data)-1),
#       (min([(cube.data * cube.unit).to(seq_with_units[:, 0, 0, 0].data[0].unit).value
#             for cube in seq_with_units[:, 0, 0, 0].data]),
#        max([(cube.data * cube.unit).to(seq_with_units[:, 0, 0, 0].data[0].unit).value
#             for cube in seq_with_units[:, 0, 0, 0].data])))),

#     (seq_with_uncertainty[:, 0, 0, 0], {},
#      (np.arange(len(seq_with_uncertainty.data)), np.array([1, 11,  1, 11]),
#       "meta.obs.sequence [None]", "Data [None]", (0, len(seq_with_uncertainty[:, 0, 0, 0].data)-1),
#       (min([cube.data for cube in seq_with_uncertainty[:, 0, 0, 0].data]),
#        max([cube.data for cube in seq_with_uncertainty[:, 0, 0, 0].data])))),

#     (seq_with_units_and_uncertainty[:, 0, 0, 0], {},
#      (np.arange(len(seq_with_units_and_uncertainty.data)), np.array([1, 0.011,  1, 0.011]),
#       "meta.obs.sequence [None]", "Data [km]",
#       (0, len(seq_with_units_and_uncertainty[:, 0, 0, 0].data)-1),
#       (min([(cube.data*cube.unit).to(seq_with_units_and_uncertainty[:, 0, 0, 0].data[0].unit).value
#             for cube in seq_with_units_and_uncertainty[:, 0, 0, 0].data]),
#        max([(cube.data*cube.unit).to(seq_with_units_and_uncertainty[:, 0, 0, 0].data[0].unit).value
#             for cube in seq_with_units_and_uncertainty[:, 0, 0, 0].data])))),

#     (seq_with_units_and_some_uncertainty[:, 0, 0, 0], {},
#      (np.arange(len(seq_with_units_and_some_uncertainty.data)), np.array([1, 0.011,  1, 0.011]),
#       "meta.obs.sequence [None]", "Data [km]",
#       (0, len(seq_with_units_and_some_uncertainty[:, 0, 0, 0].data)-1),
#       (min([(cube.data*cube.unit).to(
#           seq_with_units_and_some_uncertainty[:, 0, 0, 0].data[0].unit).value
#           for cube in seq_with_units_and_some_uncertainty[:, 0, 0, 0].data]),
#        max([(cube.data*cube.unit).to(
#            seq_with_units_and_some_uncertainty[:, 0, 0, 0].data[0].unit).value
#            for cube in seq_with_units_and_some_uncertainty[:, 0, 0, 0].data])))),

#     (seq[:, 0, 0, 0], {"axes_coordinates": "distance"},
#      ((seq.sequence_axis_extra_coords["distance"]), np.array([1, 11,  1, 11]),
#       "distance [{0}]".format(seq.sequence_axis_extra_coords["distance"].unit), "Data [None]",
#      (min(seq.sequence_axis_extra_coords["distance"].value),
#       max(seq.sequence_axis_extra_coords["distance"].value)),
#      (min([cube.data.min() for cube in seq[:, 0, 0, 0].data]),
#       max([cube.data.max() for cube in seq[:, 0, 0, 0].data])))),

#     (seq[:, 0, 0, 0], {"axes_coordinates": u.Quantity(np.arange(len(seq.data)), unit=u.cm),
#                        "axes_units": u.km},
#      (u.Quantity(np.arange(len(seq.data)), unit=u.cm).to(u.km), np.array([1, 11,  1, 11]),
#       "meta.obs.sequence [km]", "Data [None]",
#      (min((u.Quantity(np.arange(len(seq.data)), unit=u.cm).to(u.km).value)),
#       max((u.Quantity(np.arange(len(seq.data)), unit=u.cm).to(u.km).value))),
#      (min([cube.data.min() for cube in seq[:, 0, 0, 0].data]),
#       max([cube.data.max() for cube in seq[:, 0, 0, 0].data]))))
#     ])
# def test_sequence_plot_1D_plot(test_input, test_kwargs, expected_values):
#     # Unpack expected values
#     expected_x_data, expected_y_data, expected_xlabel, expected_ylabel, \
#       expected_xlim, expected_ylim = expected_values
#     # Run plot method
#     output = test_input.plot(**test_kwargs)
#     # Check values are correct
#     assert isinstance(output, matplotlib.axes.Axes)
#     np.testing.assert_array_equal(output.lines[0].get_xdata(), expected_x_data)
#     np.testing.assert_array_equal(output.lines[0].get_ydata(), expected_y_data)
#     assert output.axes.get_xlabel() == expected_xlabel
#     assert output.axes.get_ylabel() == expected_ylabel
#     output_xlim = output.axes.get_xlim()
#     assert output_xlim[0] <= expected_xlim[0]
#     assert output_xlim[1] >= expected_xlim[1]
#     output_ylim = output.axes.get_ylim()
#     assert output_ylim[0] <= expected_ylim[0]
#     assert output_ylim[1] >= expected_ylim[1]


# @pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
#     (seq[:, :, 0, 0], {},
#      (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
#                 0.49998731, 0.99989848, 0.49998731, 0.99989848]),
#       np.array([1, 2, 11, 22,  1, 2, 11, 22]),
#       "{0} [{1}]".format(seq[:, :, 0, 0].cube_like_world_axis_physical_types[common_axis], "deg"),
#       "Data [None]", tuple(seq_axis1_lim_deg),
#       (min([cube.data.min() for cube in seq[:, :, 0, 0].data]),
#        max([cube.data.max() for cube in seq[:, :, 0, 0].data])))),

#     (seq_with_units[:, :, 0, 0], {},
#      (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
#                 0.49998731, 0.99989848, 0.49998731, 0.99989848]),
#       np.array([1, 2, 0.011, 0.022,  1, 2, 0.011, 0.022]),
#       "{0} [{1}]".format(seq[:, :, 0, 0].cube_like_world_axis_physical_types[common_axis], "deg"),
#       "Data [km]", tuple(seq_axis1_lim_deg),
#       (min([min((cube.data * cube.unit).to(u.km).value)
#             for cube in seq_with_units[:, :, 0, 0].data]),
#        max([max((cube.data * cube.unit).to(u.km).value)
#             for cube in seq_with_units[:, :, 0, 0].data])))),

#     (seq_with_uncertainty[:, :, 0, 0], {},
#      (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
#                 0.49998731, 0.99989848, 0.49998731, 0.99989848]),
#       np.array([1, 2, 11, 22,  1, 2, 11, 22]),
#       "{0} [{1}]".format(
#           seq_with_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
#               common_axis], "deg"),
#       "Data [None]", tuple(seq_axis1_lim_deg),
#       (min([cube.data.min() for cube in seq_with_uncertainty[:, :, 0, 0].data]),
#        max([cube.data.max() for cube in seq_with_uncertainty[:, :, 0, 0].data])))),

#     (seq_with_some_uncertainty[:, :, 0, 0], {},
#      (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
#                 0.49998731, 0.99989848, 0.49998731, 0.99989848]),
#       np.array([1, 2, 11, 22,  1, 2, 11, 22]),
#       "{0} [{1}]".format(
#           seq_with_some_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
#               common_axis], "deg"),
#       "Data [None]", tuple(seq_axis1_lim_deg),
#       (min([cube.data.min() for cube in seq_with_some_uncertainty[:, :, 0, 0].data]),
#        max([cube.data.max() for cube in seq_with_some_uncertainty[:, :, 0, 0].data])))),

#     (seq_with_units_and_uncertainty[:, :, 0, 0], {},
#      (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
#                 0.49998731, 0.99989848, 0.49998731, 0.99989848]),
#       np.array([1, 2, 0.011, 0.022,  1, 2, 0.011, 0.022]),
#       "{0} [{1}]".format(
#           seq_with_units_and_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
#               common_axis], "deg"),
#       "Data [km]", tuple(seq_axis1_lim_deg),
#       (min([min((cube.data * cube.unit).to(u.km).value)
#             for cube in seq_with_units[:, :, 0, 0].data]),
#        max([max((cube.data * cube.unit).to(u.km).value)
#             for cube in seq_with_units[:, :, 0, 0].data])))),

#     (seq_with_units_and_some_uncertainty[:, :, 0, 0], {},
#      (np.array([0.49998731, 0.99989848, 0.49998731, 0.99989848,
#                 0.49998731, 0.99989848, 0.49998731, 0.99989848]),
#       np.array([1, 2, 0.011, 0.022,  1, 2, 0.011, 0.022]),
#       "{0} [{1}]".format(
#           seq_with_units_and_some_uncertainty[:, :, 0, 0].cube_like_world_axis_physical_types[
#               common_axis], "deg"),
#       "Data [km]", tuple(seq_axis1_lim_deg),
#       (min([min((cube.data * cube.unit).to(u.km).value)
#             for cube in seq_with_units[:, :, 0, 0].data]),
#        max([max((cube.data * cube.unit).to(u.km).value)
#             for cube in seq_with_units[:, :, 0, 0].data])))),

#     (seq[:, :, 0, 0], {"axes_coordinates": "pix"},
#      (seq[:, :, 0, 0].common_axis_extra_coords["pix"].value,
#       np.array([1, 2, 11, 22,  1, 2, 11, 22]), "pix [pix]", "Data [None]",
#       (min(seq[:, :, 0, 0].common_axis_extra_coords["pix"].value),
#        max(seq[:, :, 0, 0].common_axis_extra_coords["pix"].value)),
#       (min([cube.data.min() for cube in seq[:, :, 0, 0].data]),
#        max([cube.data.max() for cube in seq[:, :, 0, 0].data])))),

#     (seq[:, :, 0, 0],
#      {"axes_coordinates": np.arange(10, 10+seq[:, :, 0, 0].cube_like_dimensions[0].value)},
#      (np.arange(10, 10 + seq[:, :, 0, 0].cube_like_dimensions[0].value),
#       np.array([1, 2, 11, 22,  1, 2, 11, 22]),
#       "{0} [{1}]".format("", None), "Data [None]",
#       (10, 10 + seq[:, :, 0, 0].cube_like_dimensions[0].value - 1),
#       (min([cube.data.min() for cube in seq[:, :, 0, 0].data]),
#        max([cube.data.max() for cube in seq[:, :, 0, 0].data]))))
#     ])
# def test_sequence_plot_as_cube_1D_plot(test_input, test_kwargs, expected_values):
#     # Unpack expected values
#     expected_x_data, expected_y_data, expected_xlabel, expected_ylabel, \
#       expected_xlim, expected_ylim = expected_values
#     # Run plot method
#     output = test_input.plot_as_cube(**test_kwargs)
#     # Check values are correct
#     # Check type of ouput plot object
#     assert isinstance(output, matplotlib.axes.Axes)
#     # Check x and y data are correct.
#     assert np.allclose(output.lines[0].get_xdata(), expected_x_data)
#     assert np.allclose(output.lines[0].get_ydata(), expected_y_data)
#     # Check x and y axis labels are correct.
#     assert output.axes.get_xlabel() == expected_xlabel
#     assert output.axes.get_ylabel() == expected_ylabel
#     # Check all data is contained within x and y axes limits.
#     output_xlim = output.axes.get_xlim()
#     assert output_xlim[0] <= expected_xlim[0]
#     assert output_xlim[1] >= expected_xlim[1]
#     output_ylim = output.axes.get_ylim()
#     assert output_ylim[0] <= expected_ylim[0]
#     assert output_ylim[1] >= expected_ylim[1]


# def test_sequence_plot_as_cube_error():
#     with pytest.raises(TypeError):
#         seq_no_common_axis.plot_as_cube()


# @pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
#     (seq[:, :, 0, 0], {},
#      (seq_stack[:, :, 0, 0],
#       "custom:pos.helioprojective.lat [deg]", "meta.obs.sequence [None]",
#       tuple(seq_axis1_lim_deg + [0, len(seq.data)-1]))),

#     (seq_with_units[:, :, 0, 0], {},
#      (seq_stack_km[:, :, 0, 0],
#       "custom:pos.helioprojective.lat [deg]", "meta.obs.sequence [None]",
#       tuple(seq_axis1_lim_deg + [0, len(seq.data)-1]))),

#     (seq[:, :, 0, 0], {"plot_axis_indices": [0, 1]},
#      (seq_stack[:, :, 0, 0].transpose(),
#       "meta.obs.sequence [None]", "custom:pos.helioprojective.lat [deg]",
#       tuple([0, len(seq.data)-1] + seq_axis1_lim_deg))),

#     (seq[:, :, 0, 0], {"axes_coordinates": ["pix", "distance"]},
#      (seq_stack[:, :, 0, 0],
#       "pix [pix]", "distance [cm]",
#       (min(seq[0, :, 0, 0].extra_coords["pix"]["value"].value),
#        max(seq[0, :, 0, 0].extra_coords["pix"]["value"].value),
#        min(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].value),
#        max(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].value)))),
#     # This example shows weakness of current extra coord axis values on 2D plotting!
#     # Only the coordinates from the first cube are shown.

#     (seq[:, :, 0, 0], {"axes_coordinates": [np.arange(
#         10, 10+seq[:, :, 0, 0].dimensions[-1].value), "distance"], "axes_units": [None, u.m]},
#      (seq_stack[:, :, 0, 0],
#       " [None]", "distance [m]",
#       (10, 10+seq[:, :, 0, 0].dimensions[-1].value-1,
#        min(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].to(u.m).value),
#        max(seq[:, :, 0, 0].sequence_axis_extra_coords["distance"].to(u.m).value)))),

#     (seq[:, :, 0, 0], {"axes_coordinates": [np.arange(
#         10, 10+seq[:, :, 0, 0].dimensions[-1].value)*u.deg, None], "axes_units": [u.arcsec, None]},
#      (seq_stack[:, :, 0, 0],
#       " [arcsec]", "meta.obs.sequence [None]",
#       tuple(list(
#           (np.arange(10, 10+seq[:, :, 0, 0].dimensions[-1].value)*u.deg).to(u.arcsec).value) \
#           + [0, len(seq.data)-1])))
#        ])
# def test_sequence_plot_2D_image(test_input, test_kwargs, expected_values):
#     # Unpack expected values
#     expected_data, expected_xlabel, expected_ylabel, expected_extent = expected_values
#     # Run plot method
#     output = test_input.plot(**test_kwargs)
#     # Check values are correct
#     assert isinstance(output, matplotlib.axes.Axes)
#     np.testing.assert_array_equal(output.images[0].get_array(), expected_data)
#     assert output.xaxis.get_label_text() == expected_xlabel
#     assert output.yaxis.get_label_text() == expected_ylabel
#     assert np.allclose(output.images[0].get_extent(), expected_extent, rtol=1e-3)
#     # Also check x and y values?????


# @pytest.mark.parametrize("test_input, test_kwargs, expected_error", [
#     (seq[:, :, 0, 0], {"axes_coordinates": [
#         np.arange(10, 10+seq[:, :, 0, 0].dimensions[-1].value), None],
#         "axes_units": [u.m, None]}, ValueError),

#     (seq[:, :, 0, 0], {"axes_coordinates": [
#         None, np.arange(10, 10+seq[:, :, 0, 0].dimensions[0].value)],
#         "axes_units": [None, u.m]}, ValueError)
#     ])
# def test_sequence_plot_2D_image_errors(test_input, test_kwargs, expected_error):
#     with pytest.raises(expected_error):
#         output = test_input.plot(**test_kwargs)


# @pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
#     (seq[:, :, :, 0], {},
#      (seq_concat[:, :, 0],
#       "em.wl [m]", "custom:pos.helioprojective.lat [deg]",
#       tuple(seq_axis2_lim_m + seq_axis1_lim_deg))),

#     (seq_with_units[:, :, :, 0], {},
#      (seq_concat_km[:, :, 0],
#       "em.wl [m]", "custom:pos.helioprojective.lat [deg]",
#       tuple(seq_axis2_lim_m + seq_axis1_lim_deg))),

#     (seq[:, :, :, 0], {"plot_axis_indices": [0, 1],
#                        "axes_coordinates": ["pix", "hi"]},
#      (seq_concat[:, :, 0].transpose(), "pix [pix]", "hi [s]",
#       ((seq[:, :, :, 0].common_axis_extra_coords["pix"][0].value,
#         seq[:, :, :, 0].common_axis_extra_coords["pix"][-1].value,
#         seq[:, :, :, 0].data[0].extra_coords["hi"]["value"][0].value,
#         seq[:, :, :, 0].data[0].extra_coords["hi"]["value"][-1].value)))),

#     (seq[:, :, :, 0], {"axes_coordinates": [
#         np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value) * u.m,
#         np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value) * u.m]},
#      (seq_concat[:, :, 0], " [m]", " [m]",
#       (10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value-1,
#        10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value-1))),

#     (seq[:, :, :, 0], {"axes_coordinates": [
#         np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value) * u.m,
#         np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value) * u.m],
#         "axes_units": ["cm", u.cm]},
#      (seq_concat[:, :, 0], " [cm]", " [cm]",
#       (10*100, (10+seq[:, :, :, 0].cube_like_dimensions[-1].value-1)*100,
#        10*100, (10+seq[:, :, :, 0].cube_like_dimensions[0].value-1)*100)))
#     ])
# def test_sequence_plot_as_cube_2D_image(test_input, test_kwargs, expected_values):
#     # Unpack expected values
#     expected_data, expected_xlabel, expected_ylabel, expected_extent = expected_values
#     # Run plot method
#     output = test_input.plot_as_cube(**test_kwargs)
#     # Check values are correct
#     assert isinstance(output, matplotlib.axes.Axes)
#     np.testing.assert_array_equal(output.images[0].get_array(), expected_data)
#     assert output.xaxis.get_label_text() == expected_xlabel
#     assert output.yaxis.get_label_text() == expected_ylabel
#     assert np.allclose(output.images[0].get_extent(), expected_extent, rtol=1e-3)
#     # Also check x and y values?????


# @pytest.mark.parametrize("test_input, test_kwargs, expected_error", [
#     (seq[:, :, :, 0], {"axes_coordinates": [
#         np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[-1].value), None],
#         "axes_units": [u.m, None]}, ValueError),

#     (seq[:, :, :, 0], {"axes_coordinates": [
#         None, np.arange(10, 10+seq[:, :, :, 0].cube_like_dimensions[0].value)],
#         "axes_units": [None, u.m]}, ValueError)
#     ])
# def test_sequence_plot_as_cube_2D_image_errors(test_input, test_kwargs, expected_error):
#     with pytest.raises(expected_error):
#         output = test_input.plot_as_cube(**test_kwargs)


# @pytest.mark.parametrize("test_input, test_kwargs, expected_data", [
#     (seq, {}, seq_stack.reshape(4, 1, 2, 3, 4)),
#     (seq_with_units, {}, seq_stack_km.reshape(4, 1, 2, 3, 4))
#     ])
# def test_sequence_plot_ImageAnimator(test_input, test_kwargs, expected_data):
#     # Run plot method
#     output = test_input.plot(**test_kwargs)
#     # Check plot object properties are correct.
#     assert isinstance(output, ndcube.mixins.sequence_plotting.ImageAnimatorNDCubeSequence)
#     np.testing.assert_array_equal(output.data, expected_data)


# @pytest.mark.parametrize("test_input, test_kwargs, expected_data", [
#     (seq, {}, seq_concat.reshape(1, 8, 3, 4)),
#     (seq_with_units, {}, seq_concat_km.reshape(1, 8, 3, 4))
#     ])
# def test_sequence_plot_as_cube_ImageAnimator(test_input, test_kwargs, expected_data):
#     # Run plot method
#     output = test_input.plot_as_cube(**test_kwargs)
#     # Check plot object properties are correct.
#     assert isinstance(output, ndcube.mixins.sequence_plotting.ImageAnimatorCubeLikeNDCubeSequence)
#     np.testing.assert_array_equal(output.data, expected_data)


# @pytest.mark.parametrize("test_input, expected", [
#     ((seq_with_unit0.data, None), (None, None)),
#     ((seq_with_unit0.data, u.km), (None, None)),
#     ((seq_with_units.data, None), ([u.km, u.m, u.km, u.m], u.km)),
#     ((seq_with_units.data, u.cm), ([u.km, u.m, u.km, u.m], u.cm))])
# def test_determine_sequence_units(test_input, expected):
#     output_seq_unit, output_unit = ndcube.mixins.sequence_plotting._determine_sequence_units(
#         test_input[0], unit=test_input[1])
#     assert output_seq_unit == expected[0]
#     assert output_unit == expected[1]


# def test_determine_sequence_units():
#     with pytest.raises(ValueError):
#         output_seq_unit, output_unit = ndcube.mixins.sequence_plotting._determine_sequence_units(
#             seq.data, u.m)


# @pytest.mark.parametrize("test_input, expected", [
#     ((3, 1, "time", u.s), ([1], [None, 'time', None], [None, u.s, None])),
#     ((3, None, None, None), ([-1, -2], None, None))])
# def test_prep_axes_kwargs(test_input, expected):
#     output = ndcube.mixins.sequence_plotting._prep_axes_kwargs(*test_input)
#     for i in range(3):
#         assert output[i] == expected[i]


# @pytest.mark.parametrize("test_input, expected_error", [
#     ((3, [0, 1, 2], ["time", "pix"], u.s), ValueError),
#     ((3, 0, ["time", "pix"], u.s), ValueError),
#     ((3, 0, "time", [u.s, u.pix]), ValueError),
#     ((3, 0, 0, u.s), TypeError),
#     ((3, 0, "time", 0), TypeError)])
# def test_prep_axes_kwargs_errors(test_input, expected_error):
#     with pytest.raises(expected_error):
#         output = ndcube.mixins.sequence_plotting._prep_axes_kwargs(*test_input)


# @pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
#     (seq, {"plot_axis_indices": 3},
#      (seq_stack.data, none_axis_ranges_axis3, "time [min]", "Data [None]",
#       (none_axis_ranges_axis3[-1].min(), none_axis_ranges_axis3[-1].max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq_with_units, {"plot_axis_indices": -1, "data_unit": u.km},
#      (seq_stack_km.data, none_axis_ranges_axis3, "time [min]", "Data [km]",
#       (none_axis_ranges_axis3[-1].min(), none_axis_ranges_axis3[-1].max()),
#       (seq_stack_km.data.min(), seq_stack_km.data.max()))),

#     (seq_with_masks, {"plot_axis_indices": 0},
#      (seq_stack, none_axis_ranges_axis0, "meta.obs.sequence [None]", "Data [None]",
#       (none_axis_ranges_axis0[0].min(), none_axis_ranges_axis0[0].max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq_with_some_masks, {"plot_axis_indices": 0},
#      (seq_stack, none_axis_ranges_axis0, "meta.obs.sequence [None]", "Data [None]",
#       (none_axis_ranges_axis0[0].min(), none_axis_ranges_axis0[0].max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq, {"plot_axis_indices": 0, "axes_coordinates": "distance"},
#      (seq_stack.data, distance0_none_axis_ranges_axis0, "distance [cm]", "Data [None]",
#       (seq.sequence_axis_extra_coords["distance"].value.min(),
#        seq.sequence_axis_extra_coords["distance"].value.max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq, {"plot_axis_indices": 0, "axes_coordinates": "distance", "axes_units": "mm"},
#      (seq_stack.data, distance0_none_axis_ranges_axis0_mm, "distance [mm]", "Data [None]",
#       (seq.sequence_axis_extra_coords["distance"].to("mm").value.min(),
#        seq.sequence_axis_extra_coords["distance"].to("mm").value.max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq, {"plot_axis_indices": 0,
#            "axes_coordinates": userrangequantity_none_axis_ranges_axis0[0]*u.J},
#      (seq_stack.data, userrangequantity_none_axis_ranges_axis0, " [J]", "Data [None]",
#       (userrangequantity_none_axis_ranges_axis0[0].min(),
#        userrangequantity_none_axis_ranges_axis0[0].max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq, {"plot_axis_indices": 0, "axes_units": u.erg,
#            "axes_coordinates": userrangequantity_none_axis_ranges_axis0[0]*u.J},
#      (seq_stack.data, userrangequantity_none_axis_ranges_axis0_1e7, " [erg]", "Data [None]",
#       (userrangequantity_none_axis_ranges_axis0_1e7[0].min(),
#        userrangequantity_none_axis_ranges_axis0_1e7[0].max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq, {"plot_axis_indices": 2, "axes_coordinates": "hi"},
#      (seq_stack.data, hi2_none_axis_ranges_axis2, "hi [s]", "Data [None]",
#       (hi2_none_axis_ranges_axis2[2].min(), hi2_none_axis_ranges_axis2[2].max()),
#       (seq_stack.data.min(), seq_stack.data.max()))),

#     (seq, {"plot_axis_indices": 1, "axes_coordinates": "pix"},
#      (seq_stack.data, pix1_none_axis_ranges_axis1, "pix [pix]", "Data [None]",
#       (pix1_none_axis_ranges_axis1[1].min(), pix1_none_axis_ranges_axis1[1].max()),
#       (seq_stack.data.min(), seq_stack.data.max())))
#     ])
# def test_sequence_plot_LineAnimator(test_input, test_kwargs, expected_values):
#     # Unpack expected values
#     expected_data, expected_axis_ranges, expected_xlabel, \
#       expected_ylabel, expected_xlim, expected_ylim = expected_values
#     # Run plot method.
#     output = test_input.plot(**test_kwargs)
#     # Check right type of plot object is produced.
#     assert type(output) is ndcube.mixins.sequence_plotting.LineAnimatorNDCubeSequence
#     # Check data being plotted is correct
#     np.testing.assert_array_equal(output.data, expected_data)
#     if type(expected_data) is np.ma.core.MaskedArray:
#         np.testing.assert_array_equal(output.data.mask, expected_data.mask)
#     # Check values of axes and sliders is correct.
#     for i in range(len(output.axis_ranges)):
#         print(i)
#         assert np.allclose(output.axis_ranges[i], expected_axis_ranges[i])
#     # Check plot axis labels and limits are correct
#     assert output.xlabel == expected_xlabel
#     assert output.ylabel == expected_ylabel
#     assert output.xlim == expected_xlim
#     assert output.ylim == expected_ylim


# @pytest.mark.parametrize("test_input, test_kwargs, expected_values", [
#     (seq, {"plot_axis_indices": 2, "axes_units": u.s},
#      (seq_concat.data, cubelike_none_axis_ranges_axis2_s, "time [s]", "Data [None]",
#       (cubelike_none_axis_ranges_axis2_s[2].min(), cubelike_none_axis_ranges_axis2_s[2].max()),
#       (seq_concat.data.min(), seq_concat.data.max()))),

#     (seq, {"plot_axis_indices": 0},
#      (seq_concat.data, cubelike_none_axis_ranges_axis0,
#       "custom:pos.helioprojective.lat [deg]", "Data [None]",
#       (0, 7), (seq_concat.data.min(), seq_concat.data.max()))),

#     (seq_with_masks, {"plot_axis_indices": 0},
#      (seq_concat.data, cubelike_none_axis_ranges_axis0,
#       "custom:pos.helioprojective.lat [deg]", "Data [None]",
#       (0, 7), (seq_concat.data.min(), seq_concat.data.max()))),

#     (seq_with_some_masks, {"plot_axis_indices": -3},
#      (seq_concat.data, cubelike_none_axis_ranges_axis0,
#       "custom:pos.helioprojective.lat [deg]", "Data [None]",
#       (0, 7), (seq_concat.data.min(), seq_concat.data.max()))),

#     (seqm, {"plot_axis_indices": 0},
#      (seq_concat.data, cubelike_none_axis_ranges_axis0,
#       "custom:pos.helioprojective.lon [deg]", "Data [None]",
#       (0, 7), (seq_concat.data.min(), seq_concat.data.max())))
#     ])
# def test_sequence_plot_as_cube_LineAnimator(test_input, test_kwargs, expected_values):
#     # Unpack expected values
#     expected_data, expected_axis_ranges, expected_xlabel, \
#       expected_ylabel, expected_xlim, expected_ylim = expected_values
#     # Run plot method.
#     output = test_input.plot_as_cube(**test_kwargs)
#     # Check right type of plot object is produced.
#     assert type(output) is ndcube.mixins.sequence_plotting.LineAnimatorCubeLikeNDCubeSequence
#     # Check data being plotted is correct
#     np.testing.assert_array_equal(output.data, expected_data)
#     if type(expected_data) is np.ma.core.MaskedArray:
#         np.testing.assert_array_equal(output.data.mask, expected_data.mask)
#     # Check values of axes and sliders is correct.
#     for i in range(len(output.axis_ranges)):
#         assert np.allclose(output.axis_ranges[i], expected_axis_ranges[i])
#     # Check plot axis labels and limits are correct
#     assert output.xlabel == expected_xlabel
#     assert output.ylabel == expected_ylabel
#     assert output.xlim == expected_xlim
#     assert output.ylim == expected_ylim



import inspect
from functools import wraps
from itertools import chain

import numpy as np
from astropy.wcs.wcsapi import BaseHighLevelWCS, HighLevelWCSWrapper, SlicedLowLevelWCS

from ndcube.utils import wcs as wcs_utils

__all__ = ["sanitize_wcs", "sanitize_crop_inputs", "get_crop_item_from_points"]


def sanitize_wcs(func):
    """
    A wrapper for NDCube methods to sanitise the wcs argument.

    This decorator is only designed to be used on methods of NDCube.

    It will find the wcs argument, keyword or positional and if it is None, set
    it to `self.wcs`.
    It will then verify that the WCS has a matching number of pixel dimensions
    to the dimensionality of the array. It will finally verify that the object
    passed is a HighLevelWCS object, or an ExtraCoords object.
    """
    # This needs to be here to prevent a circular import
    from ndcube.extra_coords import ExtraCoords

    @wraps(func)
    def wcs_wrapper(*args, **kwargs):
        sig = inspect.signature(func)
        params = sig.bind(*args, **kwargs)
        wcs = params.arguments.get('wcs', None)
        self = params.arguments['self']

        if wcs is None:
            wcs = self.wcs

        if not isinstance(wcs, ExtraCoords):
            if not wcs.pixel_n_dim == self.data.ndim:
                raise ValueError("The supplied WCS must have the same number of "
                                 "pixel dimensions as the NDCube object. "
                                 "If you specified `cube.extra_coords.wcs` "
                                 "please just pass `cube.extra_coords`.")

        if not isinstance(wcs, (BaseHighLevelWCS, ExtraCoords)):
            raise TypeError("wcs argument must be a High Level WCS or an ExtraCoords object.")

        params.arguments['wcs'] = wcs

        return func(*params.args, **params.kwargs)

    return wcs_wrapper


def sanitize_crop_inputs(points, wcs):
    """Sanitize inputs to NDCube crop methods.

    First arg returned signifies whether the inputs imply that cropping
    should be performed or not.
    """
    points = list(points)
    n_points = len(points)
    n_coords = [None] * n_points
    values_are_none = [False] * n_points
    for i, point in enumerate(points):
        # Ensure each point is a list
        if isinstance(point, (tuple, list)):
            points[i] = list(point)
        else:
            points[i] = [point]
        # Record number of objects in each point.
        # Later we will ensure all points have same number of objects.
        n_coords[i] = len(points[i])
        # Confirm whether point contains at least one None entry.
        if all([coord is None for coord in points[i]]):
            values_are_none[i] = True
    # If no points contain a coord, i.e. if all entries in all points are None,
    # set no-op flag to True and exit.
    if all(values_are_none):
        return True, points, wcs
    # Not not all points are of same length, error.
    if len(set(n_coords)) != 1:
        raise ValueError("All points must have same number of coordinate objects."
                         f"Number of objects in each point: {n_coords}")
    # Import must be here to avoid circular import.
    from ndcube.extra_coords import ExtraCoords
    if isinstance(wcs, ExtraCoords):
        # Determine how many dummy axes are needed
        n_dummy_axes = len(wcs._cube_array_axes_without_extra_coords)
        if n_dummy_axes > 0:
            points = [point + [None] * n_dummy_axes for point in points]
        # Convert extra coords to WCS describing whole cube.
        wcs = wcs.cube_wcs
    # Ensure WCS is low level.
    if isinstance(wcs, BaseHighLevelWCS):
        wcs = wcs.low_level_wcs
    return False, points, wcs


def get_crop_item_from_points(points, wcs, crop_by_values):
    """
    Find slice item that crops to minimum cube in array-space containing specified world points.

    Parameters
    ----------
    points : iterable of iterables
        Each iterable represents a point in real world space.
        Each element in a point gives the real world coordinate value of the point
        in high-level coordinate objects or quantities.
        (Must be consistenly high or low level within and across points.)
        Objects must be in the order required by
        wcs.world_to_array_index/world_to_array_index_values.

    wcs : `~astropy.wcs.wcsapi.BaseHighLevelWCS`, `~astropy.wcs.wcsapi.BaseLowLevelWCS`
        The WCS to use to convert the world coordinates to array indices.

    crop_by_values : `bool`
        Denotes whether cropping is done using high-level objects or "values",
        i.e. low-level objects.

    Returns
    -------
    item : `tuple` of `slice`
        The slice item for each axis of the cube which, when applied to the cube,
        will return the minimum cube in array-index-space that contains all the
        input world points.
    """
    # Define a list of lists to hold the array indices of the points
    # where each inner list gives the index of all points for that array axis.
    combined_points_array_idx = [[]] * wcs.pixel_n_dim
    # For each point compute the corresponding array indices.
    for point in points:
        # Get the arrays axes associated with each element in point.
        if crop_by_values:
            point_inputs_array_axes = []
            for i in range(wcs.world_n_dim):
                pix_axes = np.array(
                    wcs_utils.world_axis_to_pixel_axes(i, wcs.axis_correlation_matrix))
                point_inputs_array_axes.append(tuple(
                    wcs_utils.convert_between_array_and_pixel_axes(pix_axes, wcs.pixel_n_dim)))
            point_inputs_array_axes = tuple(point_inputs_array_axes)
        else:
            point_inputs_array_axes = wcs_utils.array_indices_for_world_objects(
                HighLevelWCSWrapper(wcs))
        # Get indices of array axes which correspond to only None inputs in point
        # as well as those that correspond to a coord.
        point_indices_with_inputs = []
        array_axes_with_input = []
        for i, coord in enumerate(point):
            if coord is not None:
                point_indices_with_inputs.append(i)
                array_axes_with_input.append(point_inputs_array_axes[i])
        array_axes_with_input = set(chain.from_iterable(array_axes_with_input))
        array_axes_without_input = set(range(wcs.pixel_n_dim)) - array_axes_with_input
        # Slice out the axes that do not correspond to a coord
        # from the WCS and the input point.
        wcs_slice = np.array([slice(None)] * wcs.pixel_n_dim)
        if len(array_axes_without_input):
            wcs_slice[np.array(list(array_axes_without_input))] = 0
        sliced_wcs = SlicedLowLevelWCS(wcs, slices=tuple(wcs_slice))
        sliced_point = np.array(point, dtype=object)[np.array(point_indices_with_inputs)]
        # Derive the array indices of the input point and place each index
        # in the list corresponding to its axis.
        if crop_by_values:
            point_array_indices = sliced_wcs.world_to_array_index_values(*sliced_point)
            # If returned value is a 0-d array, convert to a length-1 tuple.
            if isinstance(point_array_indices, np.ndarray) and point_array_indices.ndim == 0:
                point_array_indices = (point_array_indices.item(),)
            else:
                # Convert from scalar arrays to scalars
                point_array_indices = tuple(a.item() for a in point_array_indices)
        else:
            point_array_indices = HighLevelWCSWrapper(sliced_wcs).world_to_array_index(
                *sliced_point)
            # If returned value is a 0-d array, convert to a length-1 tuple.
            if isinstance(point_array_indices, np.ndarray) and point_array_indices.ndim == 0:
                point_array_indices = (point_array_indices.item(),)
        for axis, index in zip(array_axes_with_input, point_array_indices):
            combined_points_array_idx[axis] = combined_points_array_idx[axis] + [index]
    # Define slice item with which to slice cube.
    item = []
    result_is_scalar = True
    for axis_indices in combined_points_array_idx:
        if axis_indices == []:
            result_is_scalar = False
            item.append(slice(None))
        else:
            min_idx = min(axis_indices)
            max_idx = max(axis_indices) + 1
            if max_idx - min_idx == 1:
                item.append(min_idx)
            else:
                item.append(slice(min_idx, max_idx))
                result_is_scalar = False
    # If item will result in a scalar cube, raise an error as this is not currently supported.
    if result_is_scalar:
        raise ValueError("Input points causes cube to be cropped to a single pixel. "
                         "This is not supported.")
    return tuple(item)





import numpy as np
from astropy.wcs.wcsapi.wrappers.base import BaseWCSWrapper

__all__ = ['ReorderedLowLevelWCS']


class ReorderedLowLevelWCS(BaseWCSWrapper):
    """
    A wrapper for a low-level WCS object that has re-ordered
    pixel and/or world axes.

    Parameters
    ----------
    wcs : `~astropy.wcs.wcsapi.BaseLowLevelWCS`
        The original WCS for which to reorder axes
    pixel_order : iterable
        The indices of the original axes in the order of the
        new WCS.
    world_order : iterable
        The indices of the original axes in the order of the
        new WCS.
    """
    def __init__(self, wcs, pixel_order, world_order):
        if sorted(pixel_order) != list(range(wcs.pixel_n_dim)):
            raise ValueError(f'pixel_order should be a permutation of {list(range(wcs.pixel_n_dim))}')
        if sorted(world_order) != list(range(wcs.world_n_dim)):
            raise ValueError(f'world_order should be a permutation of {list(range(wcs.world_n_dim))}')
        self._wcs = wcs
        self._pixel_order = pixel_order
        self._world_order = world_order
        self._pixel_order_inv = np.argsort(pixel_order)
        self._world_order_inv = np.argsort(world_order)

    @property
    def world_axis_physical_types(self):
        return [self._wcs.world_axis_physical_types[idx] for idx in self._world_order]

    @property
    def world_axis_units(self):
        return [self._wcs.world_axis_units[idx] for idx in self._world_order]

    @property
    def pixel_axis_names(self):
        return [self._wcs.pixel_axis_names[idx] for idx in self._pixel_order]

    @property
    def world_axis_names(self):
        return [self._wcs.world_axis_names[idx] for idx in self._world_order]

    def pixel_to_world_values(self, *pixel_arrays):
        pixel_arrays = [pixel_arrays[idx] for idx in self._pixel_order_inv]
        world_arrays = self._wcs.pixel_to_world_values(*pixel_arrays)
        world_arrays = [world_arrays[idx] for idx in self._world_order]
        return world_arrays

    def world_to_pixel_values(self, *world_arrays):
        world_arrays = [world_arrays[idx] for idx in self._world_order_inv]
        pixel_arrays = self._wcs.world_to_pixel_values(*world_arrays)
        pixel_arrays = [pixel_arrays[idx] for idx in self._pixel_order]
        return pixel_arrays

    @property
    def world_axis_object_components(self):
        return [self._wcs.world_axis_object_components[idx] for idx in self._world_order]

    @property
    def pixel_shape(self):
        if self._wcs.pixel_shape:
            return tuple([self._wcs.pixel_shape[idx] for idx in self._pixel_order])

    @property
    def pixel_bounds(self):
        if self._wcs.pixel_bounds:
            return tuple([self._wcs.pixel_bounds[idx] for idx in self._pixel_order])

    @property
    def axis_correlation_matrix(self):
        return self._wcs.axis_correlation_matrix[self._world_order][:, self._pixel_order]




import astropy.units as u
import pytest

from ndcube.utils import collection as collection_utils


@pytest.mark.parametrize("data_dimensions1,data_dimensions2,data_axes1,data_axes2", [
    ([3., 4., 5.]*u.pix, [3., 5., 15.]*u.pix, (0, 2), (0, 1))])
def test_assert_aligned_axes_compatible(data_dimensions1, data_dimensions2,
                                        data_axes1, data_axes2):
    collection_utils.assert_aligned_axes_compatible(data_dimensions1, data_dimensions2,
                                                    data_axes1, data_axes2)


@pytest.mark.parametrize("data_dimensions1,data_dimensions2,data_axes1,data_axes2", [
    ([3., 4., 5.]*u.pix, [3., 5., 15.]*u.pix, (0, 1), (0, 1)),
    ([3., 4., 5.]*u.pix, [3., 5., 15.]*u.pix, (0, 1), (0, 1, 2))])
def test_assert_aligned_axes_compatible_error(data_dimensions1, data_dimensions2,
                                              data_axes1, data_axes2):
    with pytest.raises(ValueError):
        collection_utils.assert_aligned_axes_compatible(data_dimensions1, data_dimensions2,
                                                        data_axes1, data_axes2)



from astropy.wcs.wcsapi import BaseLowLevelWCS
from mpl_animators import ArrayAnimatorWCS

from .base import BasePlotter
from .plotting_utils import prep_plot_kwargs

__all__ = ['MatplotlibSequencePlotter']


class MatplotlibSequencePlotter(BasePlotter):
    """
    Provide visualization methods for NDCubeSequence which use `matplotlib`.

    This plotter delegates much of the visualization to the `ndcube.NDCube.plot`
    which is assumed to employ the `~ndcube.visualization.mpl_plotter.MatplotlibPlotter`.
    """
    def plot(self, sequence_axis_coords=None, sequence_axis_unit=None, **kwargs):
        """
        Visualize the `~ndcube.NDCubeSequence`.

        Parameters
        ----------
        sequence_axis_coords: `str` or array-like (optional)
            The real world value of each step along the sequene axis.
            If `str`, the values are taken from `ndcube.NDCubeSequence.sequence_axis_coords`.

        sequence_axis_unit: `str` or `astropy.units.Unit` (optional)
            The unit in which to display the sequence_axis_coords.
        """
        sequence_dims = self._ndcube.dimensions
        if len(sequence_dims) == 2:
            raise NotImplementedError("Visualizing sequences of 1-D cubes not currently supported.")
        else:
            return self.animate(sequence_axis_coords, sequence_axis_unit, **kwargs)

    def animate(self, sequence_axis_coords=None, sequence_axis_unit=None, **kwargs):
        """
        Animate the `~ndcube.NDCubeSequence` with the sequence axis as a slider.

        **kwargs are passed to
        `ndcube.visualization.mpl_plotter.MatplotlibPlotter.plot` and therefore only
        apply to cube axes, not the sequence axis.
        See that method's docstring for definition of **kwargs.

        Parameters
        ----------
        sequence_axis_coords: `str` optional
            The name of the coordinate in `~ndcube.NDCubeSequence.sequence_axis_coords`
            to be used as the slider pixel values.
            If None, array indices will be used.

        sequence_axis_units: `astropy.units.Unit` or `str` (optional)
            The unit in which the sequence_axis_coordinates should be displayed.
            If None, the default unit will be used.
        """
        return SequenceAnimator(self._ndcube, sequence_axis_coords=None, sequence_axis_unit=None, **kwargs)


class SequenceAnimator(ArrayAnimatorWCS):
    """
    Animate an NDCubeSequence of NDCubes with >1 dimension.

    The sequence axis is always set as a sliders axis.
    All kwargs are passed to `ndcube.NDCube.plot`.
    The bulk of the plotting work is performed by `ndcube.NDCube.plot`
    which is assumed to exist and to call a matplotlib-based animator.

    Parameters
    ----------
    sequence: `~ndcube.NDCubeSequence`
        The sequence to animate.

    sequence_axis_coords: `str` or array-like (optional)
        The real world value of each step along the sequene axis.
        If `str`, the values are taken from `ndcube.NDCubeSequence.sequence_axis_coords`.

    sequence_axis_unit: `str` or `astropy.units.Unit` (optional)
        The unit in which to display the sequence_axis_coords.
    """

    def __init__(self, sequence, sequence_axis_coords=None, sequence_axis_unit=None, **kwargs):
        if sequence_axis_coords is not None:
            raise NotImplementedError("Setting sequence_axis_coords not yet supported.")
        if sequence_axis_unit is not None:
            raise NotImplementedError("Setting sequence_axis_unit not yet supported.")

        # Store sequence data
        self._cubes = sequence.data

        #  Process kwargs used by cube plotter.
        plot_axes = kwargs.pop("plot_axes", None)
        axes_coordinates = kwargs.pop("axes_coordinates", None)
        axes_units = kwargs.pop("axes_units", None)
        self._data_unit = kwargs.pop("data_unit", None)
        init_idx = 0
        n_cube_dims = len(self._cubes[init_idx].dimensions)
        init_wcs = self._cubes[init_idx].wcs
        self._plot_axes, self._axes_coordinates, self._axes_units = prep_plot_kwargs(
            n_cube_dims, init_wcs, plot_axes, axes_coordinates, axes_units)

        # Define sequence axis slider properties and add to kwargs.
        base_kwargs = {"slider_functions": [self._sequence_slider_function],
                       "slider_ranges": [[0, len(self._cubes)]]}
        base_kwargs.update(kwargs)

        # Calculate data and wcs for initial animation state and instantiate Animator.
        data, wcs, plot_axes, coord_params = self._cubes[0].plotter._prep_animate_args(
            self._cubes[0].wcs, self._plot_axes, self._axes_units, self._data_unit)
        if not isinstance(wcs, BaseLowLevelWCS):
            wcs = wcs.low_level_wcs
        super().__init__(data, wcs, plot_axes, coord_params=coord_params, **base_kwargs)

    def _sequence_slider_function(self, val, artist, slider):
        self._sequence_idx = int(val)
        self.data, self.wcs, _plot_axes, _coord_params = self._cubes[self._sequence_idx].plotter._prep_animate_args(
            self._cubes[self._sequence_idx].wcs, self._plot_axes, self._axes_units, self._data_unit)
        if self.plot_dimensionality == 1:
            self.update_plot_1d(val, artist, slider)
        elif self.plot_dimensionality == 2:
            self.update_plot_2d(val, artist, slider)



"""
ndcube
======

A base package for multi-dimensional contiguous and non-contiguous coordinate-aware arrays.

* Homepage: https://github.com/sunpy/ndcube
* Documentation: https://docs.sunpy.org/projects/ndcube/
"""
import sys

# Enforce Python version check during package import.
# Must be done before any ndcube imports
__minimum_python_version__ = "3.7"


class UnsupportedPythonError(Exception):
    """Running on an unsupported version of Python."""


if sys.version_info < tuple(int(val) for val in __minimum_python_version__.split('.')):
    # This has to be .format to keep backwards compatibly.
    raise UnsupportedPythonError(
        "sunpy does not support Python < {}".format(__minimum_python_version__))


from .extra_coords import ExtraCoords
from .global_coords import GlobalCoords
from .ndcollection import NDCollection
from .ndcube import NDCube
from .ndcube_sequence import NDCubeSequence
from .version import version as __version__

__all__ = ['NDCube', 'NDCubeSequence', "NDCollection", "ExtraCoords", "GlobalCoords"]
